#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from bokeh.io import output_notebook

from dask.diagnostics import ResourceProfiler

from IPython import get_ipython

if get_ipython():
    output_notebook()


def resource_analyze(dask_fun):
    """
    Profiles Dask operation and renders profiler output as a bokeh graph.

    >>> from ads.common.analyzer import resource_analyze
    >>> from ads.dataset.factory import DatasetFactory
    >>> @resource_analyze
    >>> def fetch_data():
    >>>     ds = DatasetFactory.open("/home/datascience/ads-examples/oracle_data/orcl_attrition.csv", target="Attrition").set_positive_class('Yes')
    >>>     return ds
    >>> fetch_data()

    E.g Output:

    A graph showing CPU and memory utilzation
    """

    def analyzer(*args, **kwargs):
        with ResourceProfiler(dt=0.25) as rProf:
            output = dask_fun(*args, **kwargs)
        rProf.visualize()
        return output

    return analyzer



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
from __future__ import annotations

import json
from typing import Any, Dict

from ads.jobs.builders.runtimes.base import Runtime


class CondaRuntime(Runtime):
    """Represents a job runtime with conda pack"""

    CONST_CONDA = "conda"
    CONST_CONDA_TYPE = "type"
    CONST_CONDA_TYPE_SERVICE = "service"
    CONST_CONDA_TYPE_CUSTOM = "published"
    CONST_CONDA_SLUG = "slug"
    CONST_CONDA_URI = "uri"
    CONST_CONDA_REGION = "region"

    @property
    def conda(self) -> dict:
        """The conda pack specification

        Returns
        -------
        dict
            A dictionary with "type" and "slug" as keys.
        """
        return self.get_spec(self.CONST_CONDA)

    def with_service_conda(self, slug: str):
        """Specifies the service conda pack for running the job

        Parameters
        ----------
        slug : str
            The slug name of the service conda pack

        Returns
        -------
        self
            The runtime instance.
        """
        return self.set_spec(
            self.CONST_CONDA,
            {
                self.CONST_CONDA_TYPE: self.CONST_CONDA_TYPE_SERVICE,
                self.CONST_CONDA_SLUG: slug
            }
        )

    def with_custom_conda(self, uri: str, region: str = None):
        """Specifies the custom conda pack for running the job

        Parameters
        ----------
        uri : str
            The OCI object storage URI for the conda pack,
            e.g. "oci://your_bucket@namespace/object_name."
            In the Environment Explorer of an OCI notebook session,
            this is shown as the "source" of the conda pack.
        region: str, optional
            The region of the bucket storing the custom conda pack, by default None.
            If region is not specified, ADS will use the region from your authentication credentials,
                * For API Key, config["region"] is used.
                * For Resource Principal, signer.region is used.
            This is required if the conda pack is stored in a different region.

        Returns
        -------
        self
            The runtime instance.

        See Also
        --------
        https://docs.oracle.com/en-us/iaas/data-science/using/conda_publishs_object.htm

        """
        conda_spec = {
            self.CONST_CONDA_TYPE: self.CONST_CONDA_TYPE_CUSTOM,
            self.CONST_CONDA_URI: uri
        }
        if region:
            conda_spec[self.CONST_CONDA_REGION] = region
        return self.set_spec(self.CONST_CONDA, conda_spec)


class ScriptRuntime(CondaRuntime):
    """Represents job runtime with scripts and conda pack"""

    CONST_ENTRYPOINT = "entrypoint"
    CONST_SCRIPT_PATH = "scriptPathURI"

    @property
    def script_uri(self) -> str:
        """The URI of the source code"""
        return self.get_spec(self.CONST_SCRIPT_PATH)

    def with_script(self, uri: str):
        """Specifies the source code script for the job

        Parameters
        ----------
        uri : str
            URI to the Python or Shell script, which can be any URI supported by fsspec,
            including http://, https:// and OCI object storage.
            For example: oci://your_bucket@your_namespace/path/to/script.py

        Returns
        -------
        self
            The runtime instance.
        """
        return self.set_spec(self.CONST_SCRIPT_PATH, uri)

    @property
    def source_uri(self) -> str:
        """The URI of the source code"""
        return self.get_spec(self.CONST_SCRIPT_PATH)

    def with_source(self, uri: str, entrypoint: str = None):
        """Specifies the source code for the job

        Parameters
        ----------
        uri : str
            URI to the source code,
            which can be a (.py/.sh) script, a zip/tar file or directory containing the scripts/modules
            If the source code is a single file, URI can be any URI supported by fsspec,
            including http://, https:// and OCI object storage.
            For example: oci://your_bucket@your_namespace/path/to/script.py
            If the source code is a directory, only local directory is supported.

        entrypoint : str, optional
            The relative path of the script to be set as entrypoint when source is a zip/tar/directory.
            By default None. This is not needed when the source is a single script.

        Returns
        -------
        self
            The runtime instance.
        """
        if entrypoint:
            self.set_spec(self.CONST_ENTRYPOINT, entrypoint)
        return self.with_script(uri)

    @property
    def entrypoint(self) -> str:
        """The relative path of the script to be set as entrypoint when source is a zip/tar/directory."""
        return self.get_spec(self.CONST_ENTRYPOINT)

    def with_entrypoint(self, entrypoint: str):
        """Specify the entrypoint for the job

        Parameters
        ----------
        entrypoint : str
            The relative path of the script to be set as entrypoint when source is a zip/tar/directory.

        Returns
        -------
        self
            The runtime instance.
        """
        return self.set_spec(self.CONST_ENTRYPOINT, entrypoint)


class PythonRuntime(ScriptRuntime):
    """Represents a job runtime with Python code
    Currently this is the same as ScriptRuntime
    """


class NotebookRuntime(CondaRuntime):
    """Represents a job runtime with Jupyter notebook"""

    CONST_NOTEBOOK_PATH = "notebookPathURI"
    EXCLUDE_TAG = "excludeTags"

    @property
    def notebook_uri(self) -> str:
        """The URI of the notebook"""
        return self.get_spec(self.CONST_NOTEBOOK_PATH)

    def with_notebook(self, path):
        """Specifies the notebook to be converted to python script and run as a job.

        Parameters
        ----------
        path : str
            The path of the Jupyter notebook

        Returns
        -------
        self
            The runtime instance.
        """
        return self.set_spec(self.CONST_NOTEBOOK_PATH, path)

    @property
    def exclude_tag(self) -> list:
        """A list of cell tags indicating cells to be excluded from the job"""
        return self.get_spec(self.EXCLUDE_TAG, [])

    def with_exclude_tag(self, exclude_tag_list):
        """Specifies the cell tags in the notebook to exclude cells from the job script.

        Parameters
        ----------
        exclude_tag_list : list
            A list of tags.

        Returns
        -------
        self
            The runtime instance.
        """
        return self.set_spec(self.EXCLUDE_TAG, exclude_tag_list)


class GitPythonRuntime(CondaRuntime):
    """Represents a job runtime with source code from git repository"""

    CONST_GIT_URL = "url"
    CONST_BRANCH = "branch"
    CONST_COMMIT = "commit"
    CONST_ENTRYPOINT = "entrypoint"
    CONST_ENTRY_FUNCTION = "entryFunction"
    CONST_OUTPUT_DIR = "outputDir"
    CONST_OUTPUT_URI = "outputUri"
    CONST_PYTHON_PATH = "pythonPath"
    CONST_GIT_SSH_SECRET_ID = "gitSecretId"
    CONST_SKIP_METADATA = "skipMetadataUpdate"

    def __init__(self, spec: Dict = None, skip_metadata_update=False) -> None:
        """Initialize Git Python Runtime.

        Parameters
        ----------
        spec : dict, optional
            Runtime specifications, by default None
        skip_metadata_update : bool, optional
            Indicate if the metadata update should be skipped after the job run, by default False.
            By default, the job run metadata will be updated with the following freeform tags:
                repo: The URL of the Git repository
                commit: The Git commit ID
                module: The entry script/module
                method: The entry function/method
                outputs. The prefix of the output files in object storage.
            This update step also requires resource principals to have the permission to update the job run.
        """
        super().__init__(spec=spec)
        if skip_metadata_update:
            self.set_spec(self.CONST_SKIP_METADATA, True)

    @property
    def skip_metadata_update(self):
        """Indicate if the metadata update should be skipped after the job run

        Returns
        -------
        bool
            True if the metadata update will be skipped. Otherwise False.
        """
        return self.get_spec(self.CONST_SKIP_METADATA, False)

    def with_source(
        self, url: str, branch: str = None, commit: str = None, secret_ocid: str = None
    ):
        """Specifies the Git repository and branch/commit for the job source code.

        Parameters
        ----------
        url : str
            URL of the Git repository.
        branch : str, optional
            Git branch name, by default None, the default branch will be used.
        commit : str, optional
            Git commit ID (SHA1 hash), by default None, the most recent commit will be used.
        secret_ocid : str
            The secret OCID storing the SSH key content for checking out the Git repository.

        Returns
        -------
        self
            The runtime instance.
        """
        self.set_spec(self.CONST_GIT_URL, url)
        self.set_spec(self.CONST_BRANCH, branch)
        self.set_spec(self.CONST_COMMIT, commit)
        self.set_spec(self.CONST_GIT_SSH_SECRET_ID, secret_ocid)
        return self

    @property
    def url(self) -> str:
        """URL of the Git repository."""
        return self.get_spec(self.CONST_GIT_URL)

    @property
    def branch(self) -> str:
        """Git branch name."""
        return self.get_spec(self.CONST_BRANCH)

    @property
    def commit(self) -> str:
        """Git commit ID (SHA1 hash)"""
        return self.get_spec(self.CONST_COMMIT)

    def with_entrypoint(self, path: str, func: str = None):
        """Specifies the entrypoint for the job.
        The entrypoint can be a script or a function in a script.

        Parameters
        ----------
        script : str
            The relative path for the script/module starting the job.
        func : str, optional
            The function name in the script for starting the job, by default None.
            If this is not specified, the script will be run with python command in a subprocess.

        Returns
        -------
        self
            The runtime instance.
        """
        self.set_spec(self.CONST_ENTRYPOINT, path)
        self.set_spec(self.CONST_ENTRY_FUNCTION, func)
        return self

    @property
    def entry_script(self) -> str:
        """The path of the entry script"""
        return self.get_spec(self.CONST_ENTRYPOINT)

    @property
    def entry_function(self) -> str:
        """The name of the entry function in the entry script"""
        return self.get_spec(self.CONST_ENTRY_FUNCTION)

    def with_output(self, output_dir, output_uri):
        """Specifies the outputs of the job.
        The outputs will be copied to the object storage bucket when the job is finished.

        Parameters
        ----------
        output_dir : str
            Output directory in the container running the job.
            The source code should write all output data into this directory.
        output_uri : str
            The OCI object storage URI prefix for saving the output files.
            For example, oci://bucket_name@namespace/path/to/directory

        Returns
        -------
        self
            The runtime instance.
        """
        self.set_spec(self.CONST_OUTPUT_DIR, output_dir)
        self.set_spec(self.CONST_OUTPUT_URI, output_uri)
        return self

    @property
    def output_dir(self) -> str:
        """Directory in the Job run container for saving output files generated in the job"""
        return self.get_spec(self.CONST_OUTPUT_DIR)

    @property
    def output_uri(self) -> str:
        """OCI object storage URI prefix for saving output files generated in the job"""
        return self.get_spec(self.CONST_OUTPUT_DIR)

    def with_python_path(self, *python_paths):
        """Specifies additional python paths for running the source code.
        The root of the Git repository is added to PYTHONPATH by default.
        Paths specified here will be appended to PYTHONPATH before appending the repository root.

        Parameters
        ----------
        python_paths :
            Additional python path(s) for running the source code.

        Returns
        -------
        self
            The runtime instance.
        """
        return self.set_spec(self.CONST_PYTHON_PATH, list(python_paths))

    @staticmethod
    def _serialize_arg(arg: Any) -> str:
        """Serialize the argument.
        This returns the argument "as is" if it is a string AND not a valid JSON payload.
        Otherwise the argument will be serialized with JSON.

        Parameters
        ----------
        arg : Any
            argument to be serialized

        Returns
        -------
        str
            Serialized argument as a string
        """
        if arg is None:
            return None
        if isinstance(arg, str):
            try:
                json.loads(arg)
            except json.JSONDecodeError:
                return arg
        return json.dumps(arg)

    def with_argument(self, *args, **kwargs):
        """Specifies the arguments for running the script/function.

        When running a python script, the arguments will be the command line arguments.
        For example, with_argument("arg1", "arg2", key1="val1", key2="val2")
        will generate the command line arguments: "arg1 arg2 --key1 val1 --key2 val2"

        When running a function, the arguments will be passed into the function.
        Arguments can also be list, dict or any JSON serializable object.
        For example, with_argument("arg1", "arg2", key1=["val1a", "val1b"], key2="val2")
        will be passed in as "your_function("arg1", "arg2", key1=["val1a", "val1b"], key2="val2")

        Returns
        -------
        self
            The runtime instance.
        """
        args = [self._serialize_arg(arg) for arg in args]
        kwargs = {k: self._serialize_arg(v) for k, v in kwargs.items()}
        return super().with_argument(*args, **kwargs)

    @property
    def python_path(self):
        """Additional python paths for running the source code."""
        return self.get_spec(self.CONST_PYTHON_PATH)

    @property
    def ssh_secret_ocid(self):
        """The OCID of the OCI Vault secret storing the Git SSH key."""
        return self.get_spec(self.CONST_GIT_SSH_SECRET_ID)


class DataFlowRuntime(Runtime):

    CONST_SCRIPT_BUCKET = "scriptBucket"
    CONST_ARCHIVE_BUCKET = "archiveBucket"
    CONST_ARCHIVE_URI = "archiveUri"
    CONST_SCRIPT_PATH = "scriptPathURI"

    def with_archive_uri(self, uri: str) -> "DataFlowRuntime":
        """
        Set archive uri (which is a zip file containing dependencies).

        Parameters
        ----------
        uri: str
            uri to the archive zip

        Returns
        -------
        DataFlowRuntime
            runtime instance itself
        """
        return self.set_spec(self.CONST_ARCHIVE_URI, uri)

    @property
    def archive_uri(self):
        """The Uri of archive zip"""
        return self.get_spec(self.CONST_ARCHIVE_URI)

    @property
    def script_uri(self) -> str:
        """The URI of the source code"""
        return self.get_spec(self.CONST_SCRIPT_PATH)

    def with_script_uri(self, path) -> "DataFlowRuntime":
        """
        Set script uri.

        Parameters
        ----------
        uri: str
            uri to the script

        Returns
        -------
        DataFlowRuntime
            runtime instance itself
        """
        return self.set_spec(self.CONST_SCRIPT_PATH, path)

    def with_script_bucket(self, bucket) -> "DataFlowRuntime":
        """
        Set object storage bucket to save the script, in case script uri given is local.

        Parameters
        ----------
        bucket: str
            name of the bucket

        Returns
        -------
        DataFlowRuntime
            runtime instance itself
        """
        return self.set_spec(self.CONST_SCRIPT_BUCKET, bucket)

    @property
    def script_bucket(self) -> str:
        """Bucket to save script"""
        return self.get_spec(self.CONST_SCRIPT_BUCKET)

    def with_archive_bucket(self, bucket) -> "DataFlowRuntime":
        """
        Set object storage bucket to save the archive zip, in case archive uri given is local.

        Parameters
        ----------
        bucket: str
            name of the bucket

        Returns
        -------
        DataFlowRuntime
            runtime instance itself
        """
        return self.set_spec(self.CONST_ARCHIVE_BUCKET, bucket)

    @property
    def archive_bucket(self) -> str:
        """Bucket to save archive zip"""
        return self.get_spec(self.CONST_ARCHIVE_BUCKET)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import os

OCI_ODSC_SERVICE_ENDPOINT = os.environ.get("OCI_ODSC_SERVICE_ENDPOINT")
OCI_IDENTITY_SERVICE_ENDPOINT = os.environ.get("OCI_IDENTITY_SERVICE_ENDPOINT")

NB_SESSION_COMPARTMENT_OCID = os.environ.get("NB_SESSION_COMPARTMENT_OCID") or None
PROJECT_OCID = os.environ.get("PROJECT_OCID") or None
NB_SESSION_OCID = os.environ.get("NB_SESSION_OCID") or None
USER_OCID = os.environ.get("USER_OCID") or None

# resource principal env vars
OCI_RESOURCE_PRINCIPAL_RPT_ENDPOINT = os.environ.get(
    "OCI_RESOURCE_PRINCIPAL_RPT_ENDPOINT"
)
OCI_RESOURCE_PRINCIPAL_VERSION = os.environ.get("OCI_RESOURCE_PRINCIPAL_VERSION")
OCI_RESOURCE_PRINCIPAL_RPT_PATH = os.environ.get("OCI_RESOURCE_PRINCIPAL_RPT_PATH")
OCI_RESOURCE_PRINCIPAL_RPT_ID = os.environ.get("OCI_RESOURCE_PRINCIPAL_RPT_ID")

TENANCY_OCID = os.environ.get("TENANCY_OCID")
OCI_REGION_METADATA = os.environ.get("OCI_REGION_METADATA")

JOB_RUN_OCID = os.environ.get("JOB_RUN_OCID") or None



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
from abc import ABC, abstractmethod
from ads.jobs.schema.validator import ValidatorFactory
from dataclasses import asdict, dataclass, field
from typing import Dict, List, Union

import fsspec
import json
import yaml

try:
    from yaml import CSafeDumper as dumper
    from yaml import CSafeLoader as loader
except:
    from yaml import SafeDumper as dumper
    from yaml import SafeLoader as loader


class Serializable(ABC):
    """Base class that represents a serializable item.

    Methods
    -------
    to_dict(self) -> dict
        Serializes the object into a dictionary.
    from_dict(cls, obj_dict) -> cls
        Returns an instance of the class instantiated from the dictionary provided.
    _write_to_file(s, uri, **kwargs)
        Write string s into location specified by uri
    _read_from_file(uri, **kwargs)
        Returns contents from location specified by URI
    to_json(self, uri=None, **kwargs)
        Returns object serialized as a JSON string
    from_json(cls, json_string=None, uri=None, **kwargs)
        Creates an object from JSON string provided or from URI location containing JSON string
    to_yaml(self, uri=None, **kwargs)
        Returns object serialized as a YAML string
    from_yaml(cls, yaml_string=None, uri=None, **kwargs)
        Creates an object from YAML string provided or from URI location containing YAML string
    from_string(cls, obj_string=None: str, uri=None, **kwargs)
        Creates an object from string provided or from URI location containing string
    """

    @abstractmethod
    def to_dict(self) -> dict:
        """Serializes instance of class into a dictionary"""
        pass

    @classmethod
    @abstractmethod
    def from_dict(cls, obj_dict: dict):
        """Returns an instance of the class instantiated by the dictionary provided

        Args:
            obj_dict (dict): Dictionary representation of the object
        """
        pass

    @staticmethod
    def _write_to_file(s: str, uri: str, **kwargs) -> None:
        """Write string s into location specified by uri

        Args:
            s (string): content
            uri (string): URI location to save string s
            kwargs (dict): keyword arguments to be passed into fsspec.open(). For OCI object storage, this should be config="path/to/.oci/config".
                           For other storage connections consider e.g. host, port, username, password, etc.
        """
        with fsspec.open(uri, "w", **kwargs) as f:
            f.write(s)

    @staticmethod
    def _read_from_file(uri: str, **kwargs) -> str:
        """Returns contents from location specified by URI

        Args:
            uri (string): URI location
            kwargs (dict): keyword arguments to be passed into fsspec.open(). For OCI object storage, this should be config="path/to/.oci/config".
                           For other storage connections consider e.g. host, port, username, password, etc.

        Returns:
            string: Contents in file specified by URI
        """
        with fsspec.open(uri, "r", **kwargs) as f:
            return f.read()

    def to_json(
        self, uri: str = None, encoder: callable = json.JSONEncoder, **kwargs
    ) -> str:
        """Returns object serialized as a JSON string

        Args:
            uri (string, optional): URI location to save the JSON string. Defaults to None.
            encoder (callable, optional): Encoder for custom data structures. Defaults to JSONEncoder.
            kwargs (dict): keyword arguments to be passed into fsspec.open(). For OCI object storage, this should be config="path/to/.oci/config".
                           For other storage connections consider e.g. host, port, username, password, etc.

        Returns:
            string: Serialized version of object
        """
        json_string = json.dumps(self.to_dict(), cls=encoder)
        if uri:
            self._write_to_file(s=json_string, uri=uri, **kwargs)
        return json_string

    @classmethod
    def from_json(
        cls,
        json_string: str = None,
        uri: str = None,
        decoder: callable = json.JSONDecoder,
        **kwargs
    ):
        """Creates an object from JSON string provided or from URI location containing JSON string

        Args:
            json_string (string, optional): JSON string. Defaults to None.
            uri (string, optional): URI location of file containing JSON string. Defaults to None.
            decoder (callable, optional): Custom decoder. Defaults to simple JSON decoder.
            kwargs (dict): keyword arguments to be passed into fsspec.open(). For OCI object storage, this should be config="path/to/.oci/config".
                           For other storage connections consider e.g. host, port, username, password, etc.

        Raises:
            ValueError: Raised if neither string nor uri is provided

        Returns:
            cls: Returns instance of the class
        """
        if json_string:
            return cls.from_dict(json.loads(json_string, cls=decoder))
        if uri:
            json_dict = json.loads(cls._read_from_file(uri, **kwargs), cls=decoder)
            return cls.from_dict(json_dict)
        raise ValueError("Must provide either JSON string or URI location")

    def to_yaml(self, uri: str = None, dumper: callable = dumper, **kwargs) -> str:
        """Returns object serialized as a YAML string

        Args:
            uri (string, optional): URI location to save the YAML string. Defaults to None.
            dumper (callable, optional): Custom YAML Dumper. Defaults to CDumper/SafeDumper.
            kwargs (dict): keyword arguments to be passed into fsspec.open(). For OCI object storage, this should be config="path/to/.oci/config".
                           For other storage connections consider e.g. host, port, username, password, etc.

        Returns:
            string: Serialized version of object
        """
        yaml_string = yaml.dump(self.to_dict(), Dumper=dumper)
        if uri:
            self._write_to_file(s=yaml_string, uri=uri, **kwargs)
        return yaml_string

    @classmethod
    def from_yaml(
        cls,
        yaml_string: str = None,
        uri: str = None,
        loader: callable = loader,
        **kwargs
    ):
        """Creates an object from YAML string provided or from URI location containing YAML string

        Args:
            yaml_string (string, optional): YAML string. Defaults to None.
            uri (string, optional): URI location of file containing YAML string. Defaults to None.
            loader (callable, optional): Custom YAML loader. Defaults to CLoader/SafeLoader.
            kwargs (dict): keyword arguments to be passed into fsspec.open(). For OCI object storage, this should be config="path/to/.oci/config".
                           For other storage connections consider e.g. host, port, username, password, etc.

        Raises:
            ValueError: Raised if neither string nor uri is provided

        Returns:
            cls: Returns instance of the class
        """
        if yaml_string:
            return cls.from_dict(yaml.load(yaml_string, Loader=loader))
        if uri:
            yaml_dict = yaml.load(cls._read_from_file(uri=uri, **kwargs), Loader=loader)
            return cls.from_dict(yaml_dict)
        raise ValueError("Must provide either YAML string or URI location")

    @classmethod
    def from_string(
        cls,
        obj_string: str = None,
        uri: str = None,
        loader: callable = loader,
        **kwargs
    ):
        """Creates an object from string provided or from URI location containing string

        Args:
            obj_string (str, optional): String representing the object
            uri (str, optional): URI location of file containing string. Defaults to None.
            loader (callable, optional): Custom YAML loader. Defaults to CLoader/SafeLoader.
            kwargs (dict): keyword arguments to be passed into fsspec.open(). For OCI object storage, this should be config="path/to/.oci/config".
                           For other storage connections consider e.g. host, port, username, password, etc.

        Returns:
            cls: Returns instance of the class
        """
        return cls.from_yaml(yaml_string=obj_string, uri=uri, loader=loader, **kwargs)

    def __repr__(self):
        """Returns printable version of object

        Returns:
            string: Serialized version of object as a YAML string
        """
        return self.to_yaml()



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
"""This module runs user code from git repository.
The following environment variables are used:
GIT_URL:
    URL to the Git repository.
GIT_BRANCH:
    Optional, the Git branch to checkout.
GIT_COMMIT:
    Optional, the Git commit to checkout. By default, the most recent git commit will be checked out.
CODE_DIR:
    Optional, the directory for saving the user code from Git repository. Defaults to "~/Code"
GIT_ENTRYPOINT:
    Relative path to the entry script/module file in the Git repository.
ENTRY_FUNCTION:
    Optional, function name in the entry script/module to be invoked.
    If this is not specified, the entry script will be run as Python script.
PYTHON_PATH:
    Optional, additional paths to be added to sys.path for looking up modules and packages.
    The root of the Git repository will be added by default.
    Multiple paths can be separated by os.pathsep, which is colon(:) for Linux and Mac, semicolon(;) for Windows.
OUTPUT_DIR:
    Optional, output directory to be copied to object storage.
OUTPUT_URI:
    Optional, object storage URI for saving files from the output directory.
GIT_SECRET_OCID:
    The OCID of the OCI vault secret storing the SSH key for Git commands.
SKIP_METADATA_UPDATE:
    If this variable exists, the update metadata step will be skipped.

JOB_RUN_OCID:
    The OCID of the job run. This is set by the job run.

This module requires the following packages:
oci
requests
GitPython

"""
import base64
import importlib
import json
import os
import random
import shutil
import string
import subprocess
import sys
import traceback
from typing import Optional
from urllib.parse import urlparse

import git
import oci
import requests
from oci.data_science import DataScienceClient
from oci.object_storage import ObjectStorageClient


SSH_KEY_FILE_PATH = "~/.ssh/id_rsa"


class GitManager:
    """Contains methods for fetching code from Git repository"""

    def __init__(self, repo_url: str):
        """Initialize the GitManager

        Parameters
        ----------
        repo_url : str
            The URL of the repository.

        Raises
        ------
        ValueError
            URL is not specified.
        """
        if not repo_url:
            raise ValueError("Please specify the URL of Git repository.")
        self.repo_url = repo_url
        code_dir = os.environ.get("CODE_DIR")
        # Use default directory if CODE_DIR is set to None or empty string.
        if not code_dir:
            code_dir = "~/Code/" + os.path.basename(repo_url).split(".", 1)[0]
        self.code_dir = os.path.abspath(os.path.expanduser(code_dir))

        print(f"Initializing code directory at {self.code_dir} ...")
        # Rename the existing directory if one already exists
        if os.path.exists(self.code_dir) and os.listdir(self.code_dir):
            shutil.move(
                self.code_dir,
                self.code_dir
                + "_"
                + "".join(random.choice(string.ascii_lowercase) for i in range(5)),
            )
        os.makedirs(self.code_dir, exist_ok=True)

        self.repo = None
        self.commit = None

    def fetch_repo(self):
        """Clones the Git repository."""
        print(f"Cloning Repo from {self.repo_url} ...")
        repo_url = self.repo_url
        if repo_url.startswith("git@"):
            repo_url = "ssh://" + repo_url
        host = urlparse(repo_url).hostname
        if host:
            os.makedirs(os.path.expanduser("~/.ssh"), exist_ok=True)
            subprocess.check_output(
                f"ssh-keyscan -H {host} >> ~/.ssh/known_hosts", shell=True
            )
            print(f"Added {host} to known hosts.")

        if "GIT_SECRET_OCID" in os.environ:
            ssh_key_path = os.path.expanduser(SSH_KEY_FILE_PATH)
            # subprocess.check_output(f'eval "$(ssh-agent -s)" && ssh-add "{ssh_key_path}"', shell=True)
            os.system("ssh -vT git@github.com")
            # Ignore the fingerprint checking
            ssh_cmd = f'ssh -i "{ssh_key_path}" -o "StrictHostKeyChecking=no" -o "UserKnownHostsFile=/dev/null"'
            env = {"GIT_SSH_COMMAND": ssh_cmd}
        else:
            env = None
        self.repo = git.Repo.clone_from(self.repo_url, self.code_dir, env=env)
        print(f"Cloned repo to: {self.code_dir}")
        return self

    def setup_code(self, branch: Optional[str] = None, commit: Optional[str] = None):
        """Checkouts the branch or commit of the Git repository.

        If neither branch nor commit is specified, the tip of the default branch will be used.
        If both branch and commit are specified, the commit will be used.

        Parameters
        ----------
        branch : str, optional
            The name of the branch, by default None
        commit : str, optional
            The commit ID (SHA1 hash), by default None

        """
        if commit:
            print(f"Checking out commit: {commit}")
            self.repo.git.checkout(commit)
        elif branch:
            print(f"Checking out branch: {branch}")
            self.repo.git.checkout(branch)
        else:
            print(f"Checking out the latest commit {self.repo.head.commit.hexsha}...")
        self.commit = self.repo.head.commit.hexsha
        return self


class CredentialManager:
    @staticmethod
    def init_oci_client(client_class):
        """Initializes OCI client with API key or Resource Principal.

        Parameters
        ----------
        client_class :
            The class of OCI client to be initialized.
        """
        if os.environ.get("API_KEY"):
            print(f"Initializing {client_class.__name__} with API Key...")
            client = client_class(oci.config.from_file())
        else:
            print(f"Initializing {client_class.__name__} with Resource Principal...")
            client = client_class(
                {}, signer=oci.auth.signers.get_resource_principals_signer()
            )
        return client

    @staticmethod
    def read_secret(secret_id):
        """Reads and decode the value of of a secret from OCI vault.

        Parameters
        ----------
        secret_id : str
            OCID of the secret

        Returns
        -------
        str
            The value of the secret decoded with ASCII.
        """
        secret_client = CredentialManager.init_oci_client(oci.secrets.SecretsClient)
        secret_bundle = secret_client.get_secret_bundle(secret_id)
        base64_secret_bytes = secret_bundle.data.secret_bundle_content.content.encode(
            "ascii"
        )
        base64_message_bytes = base64.b64decode(base64_secret_bytes)
        secret_content = base64_message_bytes.decode("ascii")
        return secret_content


class GitSSHKey:
    def __init__(self, secret_id) -> None:
        self.secret_id = secret_id
        self.key_file_path = os.path.expanduser(SSH_KEY_FILE_PATH)
        self.existing_git_ssh_cmd = None
        self.backup_file_path = None

    def _set_ssh_key(self):
        """Setup SSH key for Git command"""
        content = CredentialManager().read_secret(self.secret_id)
        # Add a new line to the SSH key in case the user forget to do so
        # SSH key without a new line at the end is considered "invalid format".
        if not content.endswith("\n"):
            content += "\n"
        os.makedirs(os.path.dirname(self.key_file_path), exist_ok=True)
        with open(self.key_file_path, "w") as f:
            f.write(content)
        # Set the correct permission for the SSH key.
        os.chmod(self.key_file_path, 0o600)
        print(f"SSH key saved to {self.key_file_path}")

    def _backup_ssh_key(self):
        """Backup the existing SSH key if one exists at the same file location."""
        if not os.path.exists(self.key_file_path):
            return
        suffix = (
            "."
            + "".join(random.choice(string.ascii_lowercase) for i in range(10))
            + ".bak"
        )
        self.backup_file_path = self.key_file_path + suffix
        os.rename(self.key_file_path, self.backup_file_path)

    def _restore_backup(self):
        """Restore the SSH key from backup."""
        if self.backup_file_path and os.path.exists(self.backup_file_path):
            os.rename(self.backup_file_path, self.key_file_path)

    def __enter__(self):
        """Setup the SSH key for Git command"""
        if self.secret_id:
            self._backup_ssh_key()
            self._set_ssh_key()

    def __exit__(self, tp, value, tb):
        """Remove the SSH key and reset the git command option"""
        if not self.key_file_path:
            return
        # Remove the ssh key file only if there is secret
        if self.secret_id:
            try:
                os.remove(self.key_file_path)
                self._restore_backup()
            except OSError:
                # Since this will not affect any job outputs,
                # ignore the error if the file operation failed for some reason.
                pass


class JobRunner:
    """Contains methods for running the job."""

    def __init__(self, job_run_ocid: str = ""):
        """Initialize the job runner

        Parameters
        ----------
        job_run_ocid : str, optional
            Job run OCID, by default "".
        For local testing purpose, job run OCID can be set to empty string
        """
        self.job_run_ocid = job_run_ocid
        self.git_manager = None
        self.artifacts = []
        self.source_info = {}

    def setup_python_path(self):
        """Adds additional python paths."""
        python_paths = os.environ.get("PYTHON_PATH", "").split(os.pathsep)
        python_paths.append(self.git_manager.code_dir)
        for path in python_paths:
            python_path = os.path.join(self.git_manager.code_dir, path)
            if python_path not in sys.path:
                sys.path.append(python_path)
        print(f"Python Path: {sys.path}")
        return self

    def _run_function(self, module_path: str, entry_function: str, argv: list):
        """Runs the entry function in module specified by module path.

        Parameters
        ----------
        module_path : str
            The path to the module containing the entry function.
        entry_function : str
            The name of the entry function.
        argv : list
            Argument list from command line.
        This list will be parsed into positional arguments and keyword arguments.
        """
        spec = importlib.util.spec_from_file_location("module", module_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        method = getattr(module, entry_function)

        args, kwargs = ArgumentParser(argv).parse()

        print(f"Invoking method: {entry_function} with args: {args}, kwargs: {kwargs}")
        method(*args, **kwargs)

    def _run_script(self, module_path: str, arguments: Optional[list] = None):
        """Runs the script specified by the user

        Parameters
        ----------
        module_path : str
            The path to the entry script.
        arguments : list, optional
            A list of command line arguments, by default None.
        """
        # Use -u option to run python so that the outputs will not be buffered.
        # This will allow user to see the logs sooner in long running job.
        commands = [sys.executable, "-u", module_path]
        if arguments:
            commands.extend(arguments)
        # Conda activate
        # https://docs.conda.io/projects/conda/en/latest/release-notes.html#id241
        conda_prefix = sys.executable.split("/bin/python", 1)[0]
        cmd = f"CONDA_BASE=$(conda info --base) && source $CONDA_BASE/etc/profile.d/conda.sh && conda activate {conda_prefix}; "
        cmd += " ".join(commands)
        process = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, env=os.environ.copy(), shell=True
        )
        # Steam the outputs
        while True:
            output = process.stdout.readline()
            if process.poll() is not None and output == b"":
                break
            if output:
                print(output.decode().strip(), flush=True)
        if process.returncode != 0:
            # If there is an error, exit the main process with the same return code.
            sys.exit(process.returncode)

    def fetch_code(self):
        """Gets the source code from Git repository."""
        print("Beginning Git Clone...")
        self.git_manager = (
            GitManager(os.environ.get("GIT_URL"))
            .fetch_repo()
            .setup_code(
                branch=os.environ.get("GIT_BRANCH", None),
                commit=os.environ.get("GIT_COMMIT", None),
            )
        )
        print("Code Fetch completed.")
        return self

    def run(self, argv=None):
        """Runs the job

        Parameters
        ----------
        argv : list, optional
            A list of arguments for the entry script/function, by default None

        """
        if not argv:
            argv = []

        entry_function = os.environ.get("ENTRY_FUNCTION")
        entry_script = os.environ.get("GIT_ENTRYPOINT")
        self.source_info = {
            "repo": self.git_manager.repo_url,
            "commit": self.git_manager.commit,
            "module": entry_script,
            "method": entry_function if entry_function else "",
        }

        module_path = os.path.join(self.git_manager.code_dir, entry_script)

        if entry_function:
            print(f"Running function: {entry_function} in {module_path}...")
            self._run_function(module_path, entry_function, argv)
        else:
            print(f"Running script: {module_path}")
            self._run_script(module_path, argv)
        return self

    def copy_artifacts(self, output_dir: str, output_uri: dict) -> None:
        """Copies the output files to object storage bucket.

        Parameters
        ----------
        bucket_info : dict
            Containing information for the object storage bucket.
            The dictionary should have the following keys:
            name: the name of the bucket.
            namespace: the namespace of the bucket.
            prefix: the prefix for saving the files.
            prefix is optional, defaults to Job run ID.

        output_dir : str
            Path of the output directory containing files to be copied.
        """
        output_dir = os.path.abspath(os.path.expanduser(output_dir))
        if not os.path.exists(output_dir):
            print(f"Output directory f{output_dir} not found.")
            return
        if not output_uri:
            print("OUTPUT_URI is not defined. No file is copied.")
            return
        print(f"Copying files in {output_dir} to {output_uri}...")
        parsed = urlparse(output_uri)
        bucket_name = parsed.username
        namespace = parsed.hostname
        prefix = parsed.path
        oci_os_client = CredentialManager.init_oci_client(ObjectStorageClient)

        if not prefix:
            prefix = ""
        prefix = prefix.strip("/")

        for path, _, files in os.walk(output_dir):
            for name in files:
                file_path = os.path.join(path, name)

                with open(file_path, "rb") as pkf:
                    # Get the relative path of the file to keep the directory structure
                    relative_path = os.path.relpath(file_path, output_dir)
                    if prefix:
                        file_prefix = os.path.join(prefix, relative_path)
                    else:
                        # Save file to bucket root if prefix is empty.
                        file_prefix = relative_path

                    print(
                        f"Saving {relative_path} to {bucket_name}@{namespace}/{file_prefix}"
                    )

                    oci_os_client.put_object(
                        namespace,
                        bucket_name,
                        file_prefix,
                        pkf,
                    )
                    self.artifacts.append(f"{bucket_name}@{namespace}/{file_prefix}")

    # def update_job_run_metadata_with_python_api(
    #     self, client: DataScienceClient, metadata: dict
    # ) -> None:
    #     """Updates the metadata of the job run by calling OCI Python API
    #     This method requires the OCI Python SDK to support Job APIs.
    #     This method is not used at the moment.

    #     Parameters
    #     ----------
    #     client : DataScienceClient
    #         OCI DataScienceClient
    #     metadata : dict
    #         Metadata to be saved as freeform tags.
    #     """
    #     tags = json.loads(repr(client.get_job_run(self.job_run_ocid).data)).get(
    #         "freeform_tags", {}
    #     )
    #     tags.update(metadata)
    #     data = oci.data_science.models.UpdateJobRunDetails(freeform_tags=tags)
    #     response = client.update_job_run(self.job_run_ocid, update_job_run_details=data)
    #     print(response.data)

    @staticmethod
    def _raise_for_error(response: requests.Response):
        if response.status_code >= 400:
            try:
                message = response.json().get("message", "")
            except:
                message = response.content
            error = f"{response.status_code} {response.reason} Error for {response.url}\n{message}"
            raise requests.exceptions.HTTPError(error, response=response)

    def update_job_run_metadata_with_rest_api(
        self, client: DataScienceClient, metadata: dict
    ) -> None:
        """Updates the metadata of the job run by call OCI REST API.

        Parameters
        ----------
        client : DataScienceClient
            OCI DataScienceClient
        metadata : dict
            Metadata to be saved as freeform tags.
        """
        endpoint = f"{client.base_client.endpoint}/jobRuns/{self.job_run_ocid}"
        print(f"Request endpoint: {endpoint}")
        headers = {"accept": "application/json", "content-type": "application/json"}
        signer = client.base_client.signer
        # Get the existing tags
        response = requests.get(endpoint, headers=headers, auth=signer)
        self._raise_for_error(response)

        tags = response.json().get("freeformTags", {})
        tags.update(metadata)
        body = {
            "definedTags": None,
            "displayName": None,
            "freeformTags": tags,
        }
        # Additional Headers for PUT request
        headers.update(
            {
                "expect": "100-continue",
                "opc-request-id": client.base_client.build_request_id(),
            }
        )
        response = requests.put(endpoint, headers=headers, json=body, auth=signer)
        self._raise_for_error(response)
        # print(response.json())

    def save_metadata(self) -> None:
        """Saves the metadata to job run"""
        print("Saving Metdata to job run...")
        tags = {}
        # Source info
        for key in self.source_info:
            print(f"{key} = {self.source_info[key]}")
            if self.source_info[key]:
                tags[key] = str(self.source_info[key])
        # Output files
        if self.artifacts:
            prefix = "oci://" + os.path.commonprefix(self.artifacts)
            print(f"Job Output: {prefix}")
            tags["outputs"] = prefix

        client = CredentialManager.init_oci_client(DataScienceClient)
        oci.retry.DEFAULT_RETRY_STRATEGY.make_retrying_call(
            self.update_job_run_metadata_with_rest_api, client, tags
        )
        print("Updated Job Run metadata.")


def check_internet():
    """Checks the internet connection by sending GET request to oracle.com"""
    print("Checking internet connection...")
    try:
        response = requests.get("https://oracle.com", timeout=3)
        print(f"Request Status Code: {response.status_code}")
        response.raise_for_status()
    except Exception as ex:
        print(str(ex))
        print("Internet is not available!")


class ArgumentParser:
    """Contains methods for parsing arguments for entry function."""

    def __init__(self, argument_list: list) -> None:
        """Initialize the parser with a list of arguments

        Parameters
        ----------
        argument_list : list
            A list of arguments.
        """
        self.argument_list = argument_list

    @staticmethod
    def decode_arg(val):
        """Decodes the value of the argument if it is a JSON payload.

        Parameters
        ----------
        val : str
            The argument value in a string.

        Returns
        -------
        Any
            None, if the val is None.
            String value, if the val is a string but not a JSON payload.
            Otherwise, the object after JSON decoded.
        """
        if val is None:
            return None
        try:
            return json.loads(val)
        except json.decoder.JSONDecodeError:
            return val

    @staticmethod
    def join_values(value_list):
        """Joins the values of a keyword argument.

        Parameters
        ----------
        value_list : list
            Values in a list of strings.

        Returns
        -------
        str or None
            The value of the argument as a string.
        """
        if value_list:
            return " ".join(value_list)
        return None

    def parse(self):
        """Parses the arguments

        Returns
        -------
        (list, dict)
            A tuple of positional arguments (list) and keyword arguments (dict).
        """
        args = []
        kwargs = {}
        parsing_kwargs = False
        key = None
        val = []
        for arg in self.argument_list:
            arg = str(arg)
            if len(arg) > 2 and arg.startswith("--"):
                if key:
                    # Save previous key and val
                    kwargs[key] = self.join_values(val)
                parsing_kwargs = True
                key = arg[2:]
                # Reset val
                val = []
            elif parsing_kwargs:
                val.append(arg)
            else:
                args.append(arg)
        # Save the last key and val
        if key:
            kwargs[key] = self.join_values(val)

        args = [self.decode_arg(arg) for arg in args]
        kwargs = {k: self.decode_arg(v) for k, v in kwargs.items()}
        return args, kwargs


def main():
    """The main function for running the job."""
    job_run_ocid = os.environ.get("JOB_RUN_OCID")
    print(f"Job Run ID is: {job_run_ocid}")

    check_internet()
    jbr = JobRunner(job_run_ocid)
    with GitSSHKey(os.environ.get("GIT_SECRET_OCID")):
        jbr.fetch_code().setup_python_path().run(sys.argv[1:])

    # Copy outputs
    if "OUTPUT_DIR" in os.environ:
        print(f"Found OUTPUT_DIR is configured as: {os.environ['OUTPUT_DIR']}")
        jbr.copy_artifacts(
            output_dir=os.environ["OUTPUT_DIR"],
            output_uri=os.environ.get("OUTPUT_URI"),
        )
    else:
        print("OUTPUT_DIR is not configured. Skipping copy artifacts")

    # Save metadata only if job run OCID is available
    if job_run_ocid and "SKIP_METADATA_UPDATE" not in os.environ:
        try:
            jbr.save_metadata()
        except:
            # Allow the job run to finish successfully even if the driver script failed to save the metadata.
            traceback.print_exc()

    print("Job completed.")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*--
import logging

logger = logging.getLogger(__name__)

# TODO remove
import mlx

from .mlx_global_explainer import *
from .mlx_local_explainer import *

mlx.initjs()



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import os
import functools

import nltk
from nltk import FreqDist, ngrams, pos_tag
from nltk.stem import SnowballStemmer
from nltk.tokenize import sent_tokenize, word_tokenize

from ads.feature_engineering.feature_type.adsstring.parsers.base import Parser

nltk.data.path.append(f"{os.environ['CONDA_PREFIX']}/nltk")


class NLTKParser(Parser):

    stemmers = {}  # language -> stemmer

    def __init__(self):
        if not self.language in NLTKParser.stemmers:
            NLTKParser.stemmers[self.language] = SnowballStemmer(self.language)

    @property
    @functools.lru_cache()
    def parts_of_speech(self):
        return pos_tag(word_tokenize(self.string))

    @property
    @functools.lru_cache()
    def adjectives(self):
        return [word for (word, pos) in self.parts_of_speech if pos.startswith("JJ")]

    @property
    @functools.lru_cache()
    def nouns(self):
        return [word for (word, pos) in self.parts_of_speech if pos.startswith("NN")]

    @property
    @functools.lru_cache()
    def adverbs(self):
        return [word for (word, pos) in self.parts_of_speech if pos.startswith("RB")]

    @property
    @functools.lru_cache()
    def verbs(self):
        return [word for (word, pos) in self.parts_of_speech if pos.startswith("VB")]

    @property
    @functools.lru_cache()
    def sentences(self):
        return sent_tokenize(self.string)

    @property
    @functools.lru_cache()
    def tokens(self):
        return word_tokenize(self.string)

    @property
    def words(self):
        return [word.lower() for word in self.tokens if word.isalpha()]

    @property
    @functools.lru_cache()
    def histogram(self):
        return list(FreqDist(self.words).items())

    @property
    @functools.lru_cache()
    def bigrams(self):
        return list(ngrams(self.words, 2))

    @property
    @functools.lru_cache()
    def trigrams(self):
        return list(ngrams(self.words, 3))

    @property
    @functools.lru_cache()
    def stems(self):
        return [NLTKParser.stemmers[self.language].stem(w) for w in self.words]



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
import copy
import datetime
import json
import os
import time
from typing import Dict, Optional, List
import urllib.parse

import fsspec
import oci
import ocifs
import yaml
from IPython.display import display, HTML
from oci.data_flow.models import CreateApplicationDetails, CreateRunDetails

from ads.common.auth import default_signer
from ads.common.oci_client import OCIClientFactory
from ads.common.oci_mixin import OCIModelMixin
from ads.config import OCI_REGION_METADATA
from ads.jobs.builders.infrastructure.base import Infrastructure
from ads.jobs.builders.infrastructure.utils import (
    normalize_config,
    batch_convert_case,
)
from ads.jobs.builders.runtimes.python_runtime import DataFlowRuntime


class DataFlowApp(OCIModelMixin, oci.data_flow.models.Application):
    @classmethod
    def init_client(cls, **kwargs) -> oci.data_flow.data_flow_client.DataFlowClient:
        return cls._init_client(
            client=oci.data_flow.data_flow_client.DataFlowClient, **kwargs
        )

    @property
    def client(self) -> oci.data_flow.data_flow_client.DataFlowClient:
        return super().client

    def create(self) -> "DataFlowApp":
        """
        Create a Data Flow application.

        Returns
        -------
        DataFlowApp
            a DataFlowApp instance

        """
        resp = self.client.create_application(
            self.to_oci_model(CreateApplicationDetails)
        )
        for k, v in json.loads(repr(resp.data)).items():
            setattr(self, k, v)
        return self

    def delete(self) -> None:
        """
        Delete a Data Flow application.

        Returns
        -------
        None
        """
        self.application = self.client.delete_application(self.id)
        self.lifecycle_state = oci.data_flow.models.Application.LIFECYCLE_STATE_DELETED

    def __repr__(self) -> str:
        """Displays the object as YAML."""
        return self.to_yaml()

    def to_yaml(self) -> str:
        """Serializes the object into YAML string.

        Returns
        -------
        str
            YAML stored in a string.
        """
        return yaml.safe_dump(self.to_dict())


class DataFlowRun(OCIModelMixin, oci.data_flow.models.Run):

    TERMINATED_STATES = [
        oci.data_flow.models.Run.LIFECYCLE_STATE_CANCELED,
        oci.data_flow.models.Run.LIFECYCLE_STATE_FAILED,
        oci.data_flow.models.Run.LIFECYCLE_STATE_SUCCEEDED,
    ]

    @classmethod
    def init_client(cls, **kwargs) -> oci.data_flow.data_flow_client.DataFlowClient:
        return cls._init_client(
            client=oci.data_flow.data_flow_client.DataFlowClient, **kwargs
        )

    @property
    def client(self) -> oci.data_flow.data_flow_client.DataFlowClient:
        return super().client

    def create(self) -> "DataFlowRun":
        """
        Create a Data Flow run.

        Returns
        -------
        DataFlowRun
            a DataFlowRun instance
        """
        resp = self.client.create_run(self.to_oci_model(CreateRunDetails))
        for k, v in json.loads(repr(resp.data)).items():
            setattr(self, k, v)
        return self

    @property
    def status(self) -> str:
        """
        Show status (lifecycle state) of a run.

        Returns
        -------
        str
            status of the run
        """
        if self.id:
            resp = self.client.get_run(self.id)
            self.lifecycle_state = resp.data.lifecycle_state
        return self.lifecycle_state

    @property
    def logs(self) -> "DataFlowLogs":
        """
        Show logs from a run.
        There are three types of logs: application log, driver log and executor log,
        each with stdout and stderr separately.
        To access each type of logs,
        >>> dfr.logs.application.stdout
        >>> dfr.logs.driver.stderr

        Returns
        -------
        DataFlowLogs
            an instance of DataFlowLogs
        """
        return DataFlowLogs(run_id=self.id)

    def wait(self, interval: int = 3) -> "DataFlowRun":
        """
        Wait for a run to terminate.

        Parameters
        ----------
        interval: int, optional
            interval to wait before probing again

        Returns
        -------
        DataFlowRun
            a DataFlowRun instance
        """
        current = self.status
        timestamp = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
        print(f"{timestamp} - {self.status}")
        while current not in self.TERMINATED_STATES:
            time.sleep(interval)
            if self.status != current:
                timestamp = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
                print(f"{timestamp} - {self.status}")
                current = self.status
        return self

    def __repr__(self) -> str:
        """Displays the object as YAML."""
        return self.to_yaml()

    def to_yaml(self) -> str:
        """Serializes the object into YAML string.

        Returns
        -------
        str
            YAML stored in a string.
        """
        run = self.to_dict()
        run["lifecycleState"] = self.status
        return yaml.safe_dump(run)

    def delete(self) -> None:
        """
        Cancel a Data Flow run if it is not yet terminated.

        Returns
        -------
        None
        """
        if self.status not in self.TERMINATED_STATES:
            self.client.delete_run(self.id)
            self.lifecycle_state = oci.data_flow.models.Run.LIFECYCLE_STATE_CANCELING

    @property
    def run_details_link(self):
        """
        Link to run details page in OCI console

        Returns
        -------
        DisplayHandle
            html display
        """
        signer = default_signer()
        if "region" in signer["config"]:
            region = signer["config"]["region"]
        else:
            region = json.loads(OCI_REGION_METADATA)["regionIdentifier"]
        return (
            f"https://console.{region}.oraclecloud.com/data-flow/runs/details/{self.id}"
        )


class _DataFlowLog:
    def __init__(self, run_id, src_type):
        self.run_id = run_id
        self.src_type = src_type

    @property
    def client(self):
        return OCIClientFactory(**default_signer()).dataflow

    @property
    def stdout(self):
        return self._get_logs("STDOUT")

    @property
    def stderr(self):
        return self._get_logs("STDERR")

    def _get_logs(self, out_type):
        logs = self.client.list_run_logs(self.run_id).data
        out = ""
        for log in logs:
            if log.type == out_type and self.src_type == log.source:
                out += (
                    self.client.get_run_log(self.run_id, log.name)
                    .data.text.lstrip("\x00")
                    .rstrip("\n")
                )
                out += "\n"
        out = out.rstrip("\n")
        return out


class DataFlowLogs:
    def __init__(self, run_id):
        self.run_id = run_id

    @property
    def application(self):
        return _DataFlowLog(run_id=self.run_id, src_type="APPLICATION")

    @property
    def driver(self):
        return _DataFlowLog(run_id=self.run_id, src_type="DRIVER")

    @property
    def executor(self):
        return _DataFlowLog(run_id=self.run_id, src_type="EXECUTOR")


class DataFlow(Infrastructure):
    def __init__(self, spec: dict = None):
        defaults = self._load_default_properties()
        if spec is None:
            super(DataFlow, self).__init__(defaults)
        else:
            spec = {k: v for k, v in spec.items() if f"with_{k}" in self.__dir__()}
            defaults.update(spec)
            super(DataFlow, self).__init__(spec)
        self.df_app = DataFlowApp(**self._spec)
        self.runtime = None
        self._name = None

    @staticmethod
    def _load_default_properties() -> dict:
        """
        Load default properties from environment variables, notebook session, etc.

        Returns
        -------
        dict
            a dictionary of default properties
        """
        allowed_shapes = [
            "VM.Standard2.1",
            "VM.Standard2.2",
            "VM.Standard2.4",
            "VM.Standard2.8",
            "VM.Standard2.16",
        ]
        defaults = {}
        if "NB_SESSION_COMPARTMENT_OCID" in os.environ:
            defaults["compartment_id"] = os.environ["NB_SESSION_COMPARTMENT_OCID"]
        if "NB_SESSION_OCID" in os.environ:
            dsc_client = OCIClientFactory(**default_signer()).data_science
            nb_session = dsc_client.get_notebook_session(
                os.environ["NB_SESSION_OCID"]
            ).data
            nb_config = nb_session.notebook_session_configuration_details
            if nb_config.shape in allowed_shapes:
                defaults["driver_shape"] = nb_config.shape
                defaults["executor_shape"] = nb_config.shape
        defaults["language"] = "PYTHON"
        defaults["spark_version"] = "2.4.4"
        defaults["num_executors"] = 1
        return defaults

    @property
    def name(self) -> str:
        """Display name of the job"""
        return self._name

    @name.setter
    def name(self, value: str):
        """Sets the display name of the job

        Parameters
        ----------
        value : str
            The display name of the job
        """
        if self.df_app:
            self.df_app.display_name = value
        self._name = value

    @property
    def job_id(self) -> Optional[str]:
        """The OCID of the job"""
        return self.get_spec("id")

    def with_compartment_id(self, id: str) -> "DataFlow":
        """
        Set compartment id for a Data Flow job.

        Parameters
        ----------
        id: str
            compartment id

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec("compartment_id", id)

    def with_configuration(self, configs: dict) -> "DataFlow":
        """
        Set configuration for a Data Flow job.

        Parameters
        ----------
        configs: dict
            dictionary of configurations

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec("configuration", configs)

    def with_driver_shape(self, shape: str) -> "DataFlow":
        """
        Set driver shape for a Data Flow job.

        Parameters
        ----------
        shape: str
            driver shape

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec("driver_shape", shape)

    def with_executor_shape(self, shape: str) -> "DataFlow":
        """
        Set executor shape for a Data Flow job.

        Parameters
        ----------
        shape: str
            executor shape

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec("executor_shape", shape)

    def with_language(self, lang: str) -> "DataFlow":
        """
        Set language for a Data Flow job.

        Parameters
        ----------
        lang: str
            language for the job

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec("language", lang)

    def with_metastore_id(self, id: str) -> "DataFlow":
        """
        Set Hive metastore id for a Data Flow job.

        Parameters
        ----------
        id: str
            metastore id

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """

        return self.set_spec("metastore_id", id)

    def with_logs_bucket_uri(self, uri: str) -> "DataFlow":
        """
        Set logs bucket uri for a Data Flow job.

        Parameters
        ----------
        uri: str
            uri to logs bucket

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec("logs_bucket_uri", uri)

    def with_num_executors(self, n: int) -> "DataFlow":
        """
        Set number of executors for a Data Flow job.

        Parameters
        ----------
        n: int
            number of executors

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec("num_executors", n)

    def with_spark_version(self, ver: str) -> "DataFlow":
        """
        Set spark version for a Data Flow job.
        Currently supported versions are 2.4.4 and 3.0.2
        Documentation: https://docs.oracle.com/en-us/iaas/data-flow/using/dfs_getting_started.htm#before_you_begin

        Parameters
        ----------
        ver: str
            spark version

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec("spark_version", ver)

    def with_warehouse_bucket_uri(self, uri: str) -> "DataFlow":
        """
        Set warehouse bucket uri for a Data Flow job.

        Parameters
        ----------
        uri: str
            uri to warehouse bucket

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec("warehouse_bucket_uri", uri)

    def with_id(self, id: str) -> "DataFlow":
        """
        Set id for a Data Flow job.

        Parameters
        ----------
        id: str
            id of a job

        Returns
        -------
        DataFlow
            the Data Flow instance itself
        """
        return self.set_spec("id", id)

    def __getattr__(self, item):
        if f"with_{item}" in self.__dir__():
            return self.get_spec(item)
        raise AttributeError(f"Attribute {item} not found.")

    def create(self, runtime: DataFlowRuntime, **kwargs) -> "DataFlow":
        """
        Create a Data Flow job given a runtime.

        Parameters
        ----------
        runtime
            runtime to bind to the Data Flow job
        kwargs
            additional keyword arguments

        Returns
        -------
        DataFlow
            a Data Flow job instance
        """
        if not self.name:
            timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M")
            self.name = f"dataflow-{timestamp}"
        payload = copy.deepcopy(self._spec)
        if not runtime.script_uri:
            raise ValueError("script uri must be specified in runtime.")
        if runtime.script_uri.split(":")[0] != "oci":
            if runtime.script_bucket:
                runtime.with_script_uri(
                    self._upload_file(runtime.script_uri, runtime.script_bucket)
                )
            else:
                raise ValueError(
                    "script bucket must be specified if script uri given is local."
                )
        if runtime.archive_uri and runtime.archive_uri.split(":")[0] != "oci":
            if runtime.archive_bucket:
                runtime.with_archive_uri(
                    self._upload_file(runtime.archive_uri, runtime.archive_bucket)
                )
            else:
                raise ValueError(
                    "archive bucket must be specified if archive uri given is local."
                )
        payload.update(
            {
                "display_name": self.name,
                "file_uri": runtime.script_uri,
                "freeform_tags": runtime.freeform_tags,
                "archive_uri": runtime.archive_uri,
            }
        )
        if not payload.get("compartment_id", None):
            raise ValueError(
                "Compartment id is required. Please specify compartment id via 'with_compartment_id()'."
            )
        self.df_app = DataFlowApp(**payload).create()
        self.with_id(self.df_app.id)
        return self

    @staticmethod
    def _upload_file(local_path, bucket):
        signer = default_signer()
        os_client = OCIClientFactory(**signer).object_storage
        namespace = os_client.get_namespace().data
        parsed = urllib.parse.urlparse(bucket)
        if parsed.scheme == "oci":
            dst_path = f"{bucket}/{os.path.basename(local_path)}"
        else:
            dst_path = f"oci://{bucket}@{namespace}/{os.path.basename(local_path)}"
        with fsspec.open(dst_path, mode="wb", **signer) as fto:
            with fsspec.open(local_path, mode="rb") as ffrom:
                fto.write(ffrom.read())
        return dst_path

    def run(
        self,
        name: str = None,
        args: List[str] = None,
        env_vars: Dict[str, str] = None,
        freeform_tags: Dict[str, str] = None,
        wait: bool = False,
        **kwargs,
    ) -> DataFlowRun:
        """
        Run a Data Flow job.

        Parameters
        ----------
        name: str, optional
            name of the run
        args: List[str], optional
            list of command line arguments
        env_vars: Dict[str, str], optional
            dictionary of environment variables (not used for data flow)
        freeform_tags: Dict[str, str], optional
            freeform tags
        wait: bool, optional
            whether to wait for a run to terminate
        kwargs
            additional keyword arguments

        Returns
        -------
        DataFlowRun
            a DataFlowRun instance
        """
        payload = kwargs
        if "application_id" not in payload:
            payload["application_id"] = self.id
        if "compartment_id" not in payload:
            payload["compartment_id"] = self.get_spec("compartment_id")
        payload["display_name"] = (
            name
            if name
            else f"{self.name}-run-{datetime.datetime.now().strftime('%Y%m%d-%H%M')}"
        )
        payload["arguments"] = args if args and len(args) > 0 else None
        payload["freeform_tags"] = freeform_tags
        payload.pop("spark_version", None)
        run = DataFlowRun(**payload).create()
        if wait:
            interval = kwargs["interval"] if "interval" in kwargs else 3
            run.wait(interval)
        return run

    def run_list(self, **kwargs) -> List[DataFlowRun]:
        """
        List runs associated with a Data Flow job.

        Parameters
        ----------
        kwargs
         additional arguments for filtering runs.

        Returns
        -------
        List[DataFlowRun]
            list of DataFlowRun instances
        """
        items = oci.pagination.list_call_get_all_results(
            self.df_app.client.list_runs,
            self.compartment_id,
            application_id=self.id,
            **kwargs,
        ).data
        return [DataFlowRun.from_oci_model(item) for item in items]

    def delete(self):
        """
        Delete a Data Flow job and canceling associated runs.

        Returns
        -------
        None
        """
        runs = self.run_list()
        for run in runs:
            run.delete()
        self.df_app.delete()

    @classmethod
    def from_id(cls, id: str) -> "DataFlow":
        """
        Load a Data Flow job given an id.

        Parameters
        ----------
        id: str
            id of the Data Flow job to load

        Returns
        -------
        DataFlow
            a Data Flow job instance
        """
        payload = batch_convert_case(
            yaml.safe_load(repr(DataFlowApp.from_ocid(id))), "snake"
        )
        config = normalize_config(payload)
        name = config.pop("display_name")
        ins = cls(config)
        rt = DataFlowRuntime()
        for k, v in config.items():
            if hasattr(DataFlowRuntime, f"with_{k}"):
                getattr(rt, f"with_{k}")(v)
        ins.runtime = rt
        ins.name = name
        return ins

    @classmethod
    def list_jobs(cls, compartment_id: str = None, **kwargs) -> List["DataFlow"]:
        """
        List Data Flow jobs in a given compartment.

        Parameters
        ----------
        compartment_id: str
            id of that compartment
        kwargs
            additional keyword arguments for filtering jobs

        Returns
        -------
        List[DataFlow]
            list of Data Flow jobs
        """
        return [
            cls.from_id(job.id)
            for job in DataFlowApp.list_resource(compartment_id, **kwargs)
        ]

    def to_dict(self) -> dict:
        """
        Serialize job to a dictionary.

        Returns
        -------
        dict
            serialized job as a dictionary
        """
        return {
            "kind": self.kind,
            "type": self.type,
            "spec": batch_convert_case(self._spec, "camel"),
        }

    @classmethod
    def from_dict(cls, config: dict) -> "DataFlow":
        """
        Load a Data Flow job instance from a dictionary of configurations.

        Parameters
        ----------
        config: dict
            dictionary of configurations

        Returns
        -------
        DataFlow
            a Data Flow job instance
        """
        return cls(spec=batch_convert_case(config["spec"], "snake"))

    def __repr__(self) -> str:
        """Displays the object as YAML."""
        return self.to_yaml()

    def to_yaml(self) -> str:
        """Serializes the object into YAML string.

        Returns
        -------
        str
            YAML stored in a string.
        """
        return yaml.safe_dump(self.to_dict())



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from .model_deployment import ModelDeployment
from .model_deployer import ModelDeployer
from .model_deployment_properties import ModelDeploymentProperties



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from typing import List, Optional

import matplotlib.pyplot as plt
from optuna.importance._base import BaseImportanceEvaluator
from optuna.study import Study
from optuna.visualization.matplotlib import plot_param_importances


def _plot_param_importances(
    study: Study,
    fig_size: tuple,
    evaluator: Optional[BaseImportanceEvaluator] = None,
    params: Optional[List[str]] = None,
):

    plot_param_importances(study, evaluator, params)
    plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import matplotlib.pyplot as plt
from optuna.study import Study
from optuna.visualization.matplotlib import plot_edf


def _get_edf_plot(study: Study, fig_size: tuple):

    plot_edf(study)
    plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from typing import Any, Dict, List, Optional

import matplotlib.pyplot as plt
from optuna.study import Study
from optuna.visualization.matplotlib import plot_contour


def _get_contour_plot(
    study: Study, fig_size: tuple, params: Optional[List[str]] = None
):
    plot_contour(study=study, params=params)
    plt.show()



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from __future__ import print_function, absolute_import

import abc

import six

@six.add_metaclass(abc.ABCMeta)
class AbstractTypeDiscoveryDetector:
    @abc.abstractmethod
    def discover(self, name, series):
        return


@six.add_metaclass(abc.ABCMeta)
class DiscreteDiscoveryDetector(AbstractTypeDiscoveryDetector):
    @abc.abstractmethod
    def discover(self, name, series):
        return



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


# Standard library
from typing import Union, List, Optional
import os

# Third party
import pandas as pd

# Oracle
from .common import utils
from .common.utils import OCIClientManager, State
import oci

from .model_deployment import ModelDeployment, DEFAULT_WAIT_TIME, DEFAULT_POLL_INTERVAL
from .model_deployment_properties import ModelDeploymentProperties

"""
APIs to interact with Oracle's Model Deployment service.

There are three main classes: ModelDeployment, ModelDeploymentDetails, ModelDeployer

One creates a ModelDeployment and deploys it under the umbrella of the ModelDeployer class. This way
multiple ModelDeployments can be unified with one ModelDeployer. The ModelDeployer class also serves
as the interface to all the deployments. ModelDeploymentDetails holds information about the particular
details of a particular deployment, such as how many instances, etc. In this way multiple, independent
ModelDeployments with the same details can be created using the ModelDeployer class.

Usage:
    >> from model_deploy.model_deployer import ModelDeployer, ModelDeploymentDetails
    >> deployer = ModelDeployer("model_dep_conf.yaml")
    >> deployment_properties = ModelDeploymentProperties('ocid1.datasciencemodel.ocn.reg.xxxxxxxxxxxxxxxxxxxxxxxxx')\
                            .with_prop('display_name', "My model display name")\
                            .with_prop("project_id", project_id)\
                            .with_prop("compartment_id", compartment_id)\
                            .with_instance_configuration(config={"INSTANCE_SHAPE":"VM.Standard2.1", "INSTANCE_COUNT":"1", 'bandwidth_mbps':10}).build()
    >> deployment_info = deployer.deploy(deployment_properties, max_wait_time=600, poll_interval=15)
    >> print(deployment_info.model_deployment_id)
    >> print(deployment_info.workflow_req_id)
    >> print(deployment_info.url)
    >> deployer.list_deployments() # Optionally pass in a status
"""

# Standard library
import os
import json
from typing import Union, Dict

# Third party
import pandas as pd

# Oracle
from .common import utils
from .common.utils import OCIClientManager, State
import oci

from .model_deployment import ModelDeployment, DEFAULT_WAIT_TIME, DEFAULT_POLL_INTERVAL
from .model_deployment_properties import ModelDeploymentProperties


class ModelDeployer:
    """ModelDeployer is the class responsible for deploying the ModelDeployment

    Attributes
    ----------
    config (dict): deployment configuration
    ds_client (DataScienceClient): data science client
    ds_composite_client (DataScienceCompositeClient): composite data science client

    Methods
    -------
    load_config(config_file)
        Load configuration file for deployment.
    deploy(model_deployment_details, **kwargs)
        Deploy the model specified by `model_deployment_details`.
    get_model_deployment(model_deployment_id:str)
        Get the ModelDeployment specified by `model_deployment_id`.
    get_model_deployment_state(model_deployment_id)
        Get the state of the current deployment specified by id.
    delete(model_deployment_id, **kwargs)
        Remove the model deployment specified by the id or Model Deployment Object
    list_deployments(status)
        lists the model deployments associated with current compartment and data
        science client
    show_deployments(status)
        shows the deployments filtered by `status` in a Dataframe

    """

    def __init__(self, config=None):
        """Initializes ModelDeployer with optional config

        Args:
            config (dict or str, optional): Configurations,
                which can be a dictionary or the path of YAML containing the config.
                If None, default config will be used.

        The following keys can be used in the config:
        auth: Authentication method, must be either "resource_principal" or "api_key".
            If auth is not specified:
                1. api_key will be used if available.
                2. If api_key is not available, resource_principal will be used.
        oci_config_file: OCI API key config file location. Defaults to "~/.oci/config"
        oci_config_profile: OCI API key config profile name. Defaults to "DEFAULT"
        oci_odsc_service_endpoint: URL of the service endpoint. Defaults to OCI production endpoint.
            This is for debugging/integration testing purpose.
        """
        # self.config contains contents of yml file/dict passed in by user
        self.config = utils.load_config(config)
        self.ds_client = OCIClientManager.oci_ds_client(self.config)
        self.ds_composite_client = OCIClientManager.oci_ds_composite_client(
            self.ds_client
        )

    def deploy(
        self,
        properties: Union[ModelDeploymentProperties, Dict] = None,
        wait_for_completion: bool = True,
        max_wait_time: int = DEFAULT_WAIT_TIME,
        poll_interval: int = DEFAULT_POLL_INTERVAL,
        **kwargs,
    ) -> ModelDeployment:
        """Deploys a model.

        Args:
            properties (ModelDeploymentProperties or dict): properties to deploy the model.
                properties can be None when kwargs are used for specifying properties.
            wait_for_completion (bool, optional): Flag set for whether to wait for deployment to complete before
                proceeding. Defaults to True
            max_wait_time (int, optional): Maximum amount of time to wait in seconds (Defaults to 1200). Negative implies infinite wait time.
            poll_interval (int, optional): Poll interval in seconds (Defaults to 30).
            kwargs: Keyword arguments for initializing ModelDeploymentProperties

        Returns:
            ModelDeployment: ModelDeployment objects that was deployed
        """
        model_deployment = ModelDeployment(
            properties,
            config=self.config,
            ds_client=self.ds_client,
            ds_composite_client=self.ds_composite_client,
            **kwargs,
        )
        return model_deployment.deploy(
            wait_for_completion, max_wait_time, poll_interval
        )

    def deploy_from_model_uri(
        self,
        model_uri: str,
        properties: Union[ModelDeploymentProperties, Dict] = None,
        wait_for_completion: bool = True,
        max_wait_time: int = DEFAULT_WAIT_TIME,
        poll_interval: int = DEFAULT_POLL_INTERVAL,
        **kwargs,
    ) -> ModelDeployment:
        """Deploys a model.

        Args:
            model_uri (str): uri to model files, can be local or in cloud storage
            properties (ModelDeploymentProperties or dict): properties to deploy the model.
                properties can be None when kwargs are used for specifying properties.
            wait_for_completion (bool, optional): Flag set for whether to wait for deployment to complete before
                proceeding. Defaults to True
            max_wait_time (int, optional): Maximum amount of time to wait in seconds (Defaults to 1200). Negative implies infinite wait time.
            poll_interval (int, optional): Poll interval in seconds (Defaults to 30).
            kwargs: Keyword arguments for initializing ModelDeploymentProperties

        Returns:
            ModelDeployment: ModelDeployment object that was deployed
        """
        if properties:
            model_id = OCIClientManager._prepare_artifact(
                model_uri=model_uri, properties=properties, ds_client=self.ds_client
            )
            properties.model_deployment_configuration_details.model_configuration_details.model_id = (
                model_id
            )
        else:
            model_id = OCIClientManager._prepare_artifact(
                model_uri=model_uri, properties=kwargs, ds_client=self.ds_client
            )
            kwargs["model_id"] = model_id

        return self.deploy(
            properties,
            wait_for_completion=wait_for_completion,
            max_wait_time=max_wait_time,
            poll_interval=poll_interval,
            **kwargs,
        )

    def update(
        self,
        model_deployment_id: str,
        properties: ModelDeploymentProperties = None,
        wait_for_completion: bool = True,
        max_wait_time: int = DEFAULT_WAIT_TIME,
        poll_interval: int = DEFAULT_POLL_INTERVAL,
        **kwargs,
    ) -> ModelDeployment:
        """Updates an existing model deployment.

        Args:
            model_deployment_id (str): Model deployment OCID
            properties (ModelDeploymentProperties, optional): An instance of ModelDeploymentProperties or
                dict to initialize the ModelDeploymentProperties. Defaults to None.
            wait_for_completion (bool, optional): Flag set for whether to wait for deployment to complete before
                proceeding. Defaults to True
            max_wait_time (int, optional): Maximum amount of time to wait in seconds (Defaults to 1200).
            poll_interval (int, optional): Poll interval in seconds (Defaults to 30).
            kwargs: Updates can also be specified in keyword arguments.

        Returns:
            ModelDeployment: a ModelDeployment object

        """

        model_deployment = self.get_model_deployment(model_deployment_id)
        # Deployment properties will be refreshed within model_deployment.update() when update is done.
        return model_deployment.update(
            properties,
            wait_for_completion,
            max_wait_time=max_wait_time,
            poll_interval=poll_interval,
            **kwargs,
        )

    def get_model_deployment(self, model_deployment_id: str) -> ModelDeployment:
        """get_model_deployment returns the ModelDeployment specified by `model_deployment_id`.

        Args:
            model_deployment_id (str): the model deployment ocid

        Returns:
            [ModelDeployment]: a ModelDeployment object
        """

        try:
            oci_model_deployment_object = self.ds_client.get_model_deployment(
                model_deployment_id
            ).data
            model_deployment_object = ModelDeployment(
                oci_model_deployment_object,
                config=self.config,
                ds_client=self.ds_client,
                ds_composite_client=self.ds_composite_client,
            )
            return model_deployment_object
        except Exception as e:
            utils.get_logger().error(
                "Getting model deployment failed with error: %s", e
            )
            raise e

    def get_model_deployment_state(self, model_deployment_id: str) -> State:
        """get_model_deployment_state returns the state of the current deployment specified by id

        Args:
            model_deployment_id (str): the model deployment ocid

        Returns:
            [str]: current state of the deployment specified by model_deployment_id
        """

        model_deployment = self.get_model_deployment(model_deployment_id)
        return model_deployment.state

    def delete(
        self,
        model_deployment_id,
        wait_for_completion: bool = True,
        max_wait_time: int = DEFAULT_WAIT_TIME,
        poll_interval: int = DEFAULT_POLL_INTERVAL,
    ) -> ModelDeployment:
        """delete removes the model deployment specified by the id or Model Deployment Object

        Args:
            model_deployment (str): the model deployment ocid
            wait_for_completion (bool, optional): Wait for deletion to complete. Defaults to True.
            max_wait_time (int, optional): Maximum amount of time to wait in seconds (Defaults to 600).
                Negative implies infinite wait time.
            poll_interval (int, optional): Poll interval in seconds (Defaults to 60).

        Returns:
            ModelDeployment object that was deleted
        """
        try:
            model_deployment_object = self.get_model_deployment(model_deployment_id)
            return model_deployment_object.delete(
                wait_for_completion, max_wait_time, poll_interval
            )
        except Exception as e:
            utils.get_logger().error(
                "Deleting model deployment failed with error: %s", format(e)
            )
            raise e

    def list_deployments(self, status=None, compartment_id=None, **kwargs) -> list:
        """Lists the model deployments associated with current compartment and data science client

        Args:
            status (str, optional): Status of deployment. Defaults to None.
            compartment_id (str, optional): Target compartment to list deployments from.
                Defaults to the compartment set in the environment variable "NB_SESSION_COMPARTMENT_OCID".
                If "NB_SESSION_COMPARTMENT_OCID" is not set, the root compartment ID will be used.
                An ValueError will be raised if root compartment ID cannot be determined.
            kwargs: The values are passed to oci.data_science.DataScienceClient.list_model_deployments.

        Returns:
            list: A list of ModelDeployment objects.

        Raises:
            ValueError: If compartment_id is not specified and cannot be determined from the environment.

        """
        if not compartment_id:
            compartment_id = OCIClientManager.default_compartment_id()
        if not compartment_id:
            raise ValueError(
                "Unable to determine compartment ID from environment. Please specify compartment_id."
            )

        if isinstance(status, State):
            status = status.name

        if status is not None:
            kwargs["lifecycle_state"] = status

        # https://oracle-cloud-infrastructure-python-sdk.readthedocs.io/en/latest/api/pagination.html#module-oci.pagination
        deployments = oci.pagination.list_call_get_all_results(
            self.ds_client.list_model_deployments, compartment_id, **kwargs
        ).data

        return [ModelDeployment(deployment) for deployment in deployments]

    def show_deployments(
        self,
        status=None,
        compartment_id=os.environ.get("NB_SESSION_COMPARTMENT_OCID", None),
    ) -> pd.DataFrame:
        """Returns the model deployments associated with current compartment and data science client
            as a Dataframe that can be easily visualized

        Args:
            status ([str], optional): Status of deployment. Defaults to None.
            compartment_id ([str], optional): Target compartment from which to list deployments.
                                              Defaults to compartment set in the environment.

        Returns:
            [pd.DataFrame]: Dataframe containing information about the ModelDeployments
        """
        if type(status) == str or status == None:
            status = State._from_str(status)
        model_deployments = self.ds_client.list_model_deployments(compartment_id).data
        display = pd.DataFrame()
        ids, urls, status_list = [], [], []
        for oci_model_deployment_object in model_deployments:
            state_of_model = State._from_str(
                oci_model_deployment_object.lifecycle_state
            )
            if status == State.UNKNOWN or status.name == state_of_model.name:
                model_dep_id = oci_model_deployment_object.id
                model_dep_url = oci_model_deployment_object._model_deployment_url
                model_dep_state = oci_model_deployment_object.lifecycle_state
                ids.append(model_dep_id)
                urls.append(model_dep_url)
                status_list.append(model_dep_state)
        display["model_id"] = ids
        display["deployment_url"] = urls
        display["current_state"] = status_list
        return display



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
"""Utilities used by the model deployment package
"""

# Standard lib
import fsspec
import json
# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import logging
import oci
import os
import shutil
import sys
import tempfile
import time
import yaml
from typing import Dict

from enum import Enum, auto
from oci.data_science.models import CreateModelDetails
from oci.signer import Signer
from oci.util import get_signer_from_authentication_type, AUTHENTICATION_TYPE_FIELD_NAME
from .progress_bar import TqdmProgressBar, DummyProgressBar


SERVICE_ENDPOINT_CONFIG = "oci_odsc_service_endpoint"
logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger("ODSC-ModelDeployment")


def get_logger():
    return logger


def load_config(source=None):
    """Loads SDK config from YAML or dict

    Args:
        source (str or dict, optional): Path to config YAML or a dict containing the configurations.
            Defaults to None. Source will be ignored if it is not str or dict.

    Returns:
        dict: Dictionary containing the configurations
    """
    # Empty config dict
    config = {}

    # source will be ignore if it is not str or dict
    if isinstance(source, dict):
        # source is dict
        config = source
    elif isinstance(source, str):
        # source is path to yaml file
        if "yaml" in source:
            with open(source) as cf:
                try:
                    config = yaml.load(cf, Loader=yaml.CLoader)
                except:
                    config = yaml.load(cf, Loader=yaml.Loader)
        else:
            logger.error("Pass in a valid config yaml file")
            raise ValueError(f"Not a valid yaml file: {source}")

    # Load service endpoint from environment variable if it is not specified in config
    # Default service endpoint in OCI SDK will be used if service endpoint is not in config or environment variables.
    if not config.get(SERVICE_ENDPOINT_CONFIG) and os.environ.get(
        "OCI_ODSC_SERVICE_ENDPOINT"
    ):
        config[SERVICE_ENDPOINT_CONFIG] = os.environ["OCI_ODSC_SERVICE_ENDPOINT"]
    config = {str(k).lower(): v for k, v in config.items()}
    return config


def set_log_level(level="INFO"):
    """set_log_level sets the logger level

    Args:
        level (str, optional): The logger level. Defaults to "INFO"

    Returns:
        Nothing
    """

    level = logging.getLevelName(level)
    logger.setLevel(level)


def seconds_since(t):
    """seconds_since returns the seconds since `t`. `t` is assumed to be a time
    in epoch seconds since time.time() returns the current time in epoch seconds.

    Args:
        t (int) - a time in epoch seconds

    Returns
        int: the number of seconds since `t`
    """

    return time.time() - t


def is_notebook():
    """is_notebook returns True if the environment is a Jupyter notebook and
    False otherwise

    Args:
        None

    Returns:
        bool: True if Jupyter notebook; False otherwise

    Raises:
        NameError: If retrieving the shell name from get_ipython() throws an error

    """

    try:
        from IPython import get_ipython

        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":  # pragma: no cover
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except ImportError:
        # IPython is not installed
        return False
    except NameError:
        return False  # Probably standard Python interpreter


def get_progress_bar(max_progress, description="Initializing"):
    """get_progress_bar return an instance of ProgressBar, sensitive to the runtime environment

    Args:
        max_progress (int): Progress bar max
        description (str, optional): Progress bar description (defaults to "Initializing")

    Returns:
        An instance of ProgressBar. Either a DummyProgressBar (non-notebook) or TqdmProgressBar
        (notebook environement)
    """

    if is_notebook():  # pragma: no cover
        return TqdmProgressBar(max_progress, description=description, verbose=False)
    else:
        return DummyProgressBar()


# State Constants
class State(Enum):
    ACTIVE = auto()
    CREATING = auto()
    DELETED = auto()
    DELETING = auto()
    FAILED = auto()
    INACTIVE = auto()
    UPDATING = auto()
    UNKNOWN = auto()

    @staticmethod
    def _from_str(state):
        if state == None:
            return State.UNKNOWN
        elif state.upper() == "ACTIVE":
            return State.ACTIVE
        elif state.upper() == "CREATING":
            return State.CREATING
        elif state.upper() == "DELETED":
            return State.DELETED
        elif state.upper() == "DELETING":
            return State.DELETING
        elif state.upper() == "FAILED":
            return State.FAILED
        elif state.upper() == "INACTIVE":
            return State.INACTIVE
        elif state.upper() == "UPDATING":
            return State.UPDATING
        else:
            return State.UNKNOWN

    def __call__(self):
        # This will provide backward compatibility.
        # In previous release, ModelDeployment has state() as method instead of property
        return self


class OCIClientManager:
    """OCIClientManager is a helper class used for accessing DataScienceClient and
    DataScienceCompositeClient objects

    Attributes
    ----------
    ds_client - class attribute for data science client
    ds_composite_client - class attribute for data science composite client

    Methods
    -------
    oci_ds_client(config)
        returns the OCI DataScienceClient
    oci_ds_composite_client(ds_client)
        returns the OCI DataScienceCompositeClient
    """

    @staticmethod
    def get_oci_config_path_and_profile(sdk_config=None):
        """Determines the OCI config path and profile from SDK config

        Args:
            sdk_config (dict, optional): A dictionary containing MD SDK config. Defaults to None.

        Returns:
            dict: A dictionary containing file_path and/or profile if it is specified in the SDK config.
                This dictionary can be used as kwargs for oci.config.from_file(**kwargs)
        """
        if sdk_config is None:
            sdk_config = {}
        sdk_config = {k.lower(): v for k, v in sdk_config.items()}
        # Get API key config if signer is not obtained
        # Use default OCI config and profile when the value of oci_config_profile/oci_config_file is None or empty
        profile = sdk_config.get("oci_config_profile")
        file_path = sdk_config.get("oci_config_file")
        # Pass file_path/profile to oci.config.from_file only if there is value
        # OCI SDK will handle the default values.
        kwargs = {}
        if file_path:
            kwargs["file_path"] = file_path
        if profile:
            kwargs["profile"] = profile
        return kwargs

    @staticmethod
    def get_oci_config(sdk_config=None):
        oci_config = oci.config.from_file(
            **OCIClientManager.get_oci_config_path_and_profile(sdk_config)
        )
        return oci_config

    @staticmethod
    def get_oci_config_signer(sdk_config=None):
        """Gets OCI signer from SDK config"""
        oci_config = {}
        signer = None
        if not sdk_config:
            sdk_config = {}

        # Get resource principal signer if user specified resource principal as authentication method
        if sdk_config.get("auth") == "resource_principal":
            # EnvironmentError will be raise if resource principal is not available.
            signer = oci.auth.signers.get_resource_principals_signer()
            return oci_config, signer
        # Try to get signer from OCI config
        try:
            oci_config = OCIClientManager.get_oci_config(sdk_config)

            if AUTHENTICATION_TYPE_FIELD_NAME in oci_config:
                signer = get_signer_from_authentication_type(oci_config)
            else:
                signer = Signer.from_config(oci_config)
            return oci_config, signer
        except oci.exceptions.ConfigFileNotFound:
            if "oci_api_keys" in sdk_config or sdk_config.get("auth") == "api_key":
                # If OCI_API_KEYS are specified in config,
                #   raise error to let user know config file is not found.
                raise

        # Try to get signer from resource principal only if OCI_API_KEYS are not specified in config
        try:
            signer = oci.auth.signers.get_resource_principals_signer()
        except EnvironmentError:
            raise EnvironmentError(
                "Unable to obtain OCI credentials. Please setup API Keys."
            )

        return oci_config, signer

    @staticmethod
    def oci_ds_client(sdk_config=None):
        """Returns the OCI DataScienceClient

        Returns:
            DataScienceClient: DataScienceClient
        """
        if not sdk_config:
            sdk_config = {}

        oci_config, signer = OCIClientManager().get_oci_config_signer(sdk_config)

        kwargs = {}
        if SERVICE_ENDPOINT_CONFIG in sdk_config:
            kwargs["service_endpoint"] = sdk_config[SERVICE_ENDPOINT_CONFIG]

        ds_client = oci.data_science.DataScienceClient(
            oci_config,
            signer=signer,
            retry_strategy=oci.retry.DEFAULT_RETRY_STRATEGY,
            **kwargs,
        )
        return ds_client

    @staticmethod
    def oci_ds_composite_client(ds_client):
        """oci_ds_composite_client returns the OCI DataScienceCompositeClient

        Args:
            ds_client (DataScienceClient): OCI DataScienceClient

        Returns:
            DataScienceCompositeClient: DataScienceCompositeClient
        """
        return oci.data_science.DataScienceClientCompositeOperations(ds_client)

    @staticmethod
    def default_compartment_id(config=None):
        """Determines the default compartment OCID
        This method finds the compartment OCID from (in priority order):
        an environment variable, an API key config or a resource principal signer.

        Parameters
        ----------
        config : dict, optional
            The model deployment config, which contains the following keys:
            auth: Authentication method, must be either "resource_principal" or "api_key".
            If auth is not specified:
                1. api_key will be used if available.
                2. If api_key is not available, resource_principal will be used.
            oci_config_file: OCI API key config file location. Defaults to "~/.oci/config"
            oci_config_profile: OCI API key config profile name. Defaults to "DEFAULT"

        Returns
        -------
        str or None
            The compartment OCID if found. Otherwise None.
        """
        # Try to get compartment ID from environment variable.')
        if os.environ.get('NB_SESSION_COMPARTMENT_OCID'):
            return os.environ.get('NB_SESSION_COMPARTMENT_OCID')
        # Try to get compartment ID from OCI config, then RP signer
        # Note: we assume compartment_ids can never be: 0, False, etc.
        oci_config, signer = OCIClientManager.get_oci_config_signer(config)
        return oci_config.get("compartment_id") or oci_config.get("tenancy") or \
               getattr(signer, "tenancy_id", None)


    @staticmethod
    def _prepare_artifact(model_uri: str, properties: Dict, ds_client=None) -> str:
        """
        Prepare model artifact. Returns model ocid.

        Args:
            model_uri (str): uri to model files, can be local or in cloud storage
            properties (dict): dictionary of properties that are needed for creating a model.
            ds_client (DataScienceClient): OCI DataScienceClient

        Returns:
            str: model ocid
        """
        if not ds_client:
            ds_client = OCIClientManager.oci_ds_client()
        if properties:
            properties_dict = (
                properties
                if isinstance(properties, dict)
                else json.loads(repr(properties))
            )
        with tempfile.TemporaryDirectory() as d:
            fhandlers = fsspec.open_files(
                model_uri,
                config=OCIClientManager.get_oci_config(),
                mode="rb",
            )
            if len(fhandlers) == 0:
                raise FileNotFoundError("No files found under this path.")
            for fh in fhandlers:
                with fh as fin:
                    with open(os.path.join(d, os.path.basename(fh.path)), "wb") as fout:
                        fout.write(fin.read())
            shutil.make_archive(
                os.path.join(os.path.dirname(d), "model_files"),
                "zip",
                os.path.dirname(d),
                os.path.basename(d),
            )
            return OCIClientManager._upload_artifact(
                ds_client,
                f"{os.path.join(os.path.dirname(d), 'model_files')}.zip",
                properties_dict,
            )

    @staticmethod
    def _upload_artifact(ds_client, model_zip: str, properties: dict) -> str:
        """Uploads the model artifact to cloud storage.

        Args:
            ds_client (DataScienceClient): OCI DataScienceClient
            model_zip (str): path to model artifact zip file
            properties (dict): dictionary of properties

        Returns:
            str: model ocid
        """
        create_model_details = CreateModelDetails(
            display_name=properties.get("display_name", None),
            project_id=properties["project_id"],
            compartment_id=properties["compartment_id"],
        )

        model = ds_client.create_model(create_model_details).data
        with open(model_zip, "rb") as data:
            ds_client.create_model_artifact(
                model.id,
                data,
                content_disposition=f'attachment; filename="{model.id}.zip"',
            )
        return model.id



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
import copy
import re


def _snake_to_camel(name):
    tokens = name.split("_")
    return tokens[0] + "".join(
        x.capitalize() if not x.isupper() else x for x in tokens[1:]
    )


def _camel_to_snake(name: str) -> str:
    s = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s).lower()


def batch_convert_case(spec: dict, to_fmt: str) -> dict:
    """
    Convert the case of a dictionary of spec from camel to snake or vice versa.

    Parameters
    ----------
    spec: dict
        dictionary of spec to convert
    to_fmt: str
        format to convert to, can be "camel" or "snake"

    Returns
    -------
    dict
        dictionary of converted spec
    """
    converted = {}
    if to_fmt == "camel":
        converter = _snake_to_camel
    else:
        converter = _camel_to_snake
    for k, v in spec.items():
        if k == "spec":
            converted[converter(k)] = batch_convert_case(v, to_fmt)
        else:
            converted[converter(k)] = v
    return converted


ALIAS_MAP = {
    "file_uri": "script_uri",
}


def normalize_config(config: dict) -> dict:
    """
    Normalize property names in a configuration so that they work directly with ADS.

    Parameters
    ----------
    config: dict
        input configuration, usually coming directly from an OCI response

    Returns
    -------
    dict
        output configuration
    """
    normalized = {}
    for k, v in config.items():
        k = ALIAS_MAP[k] if k in ALIAS_MAP else k
        if isinstance(v, dict):
            normalized[k] = normalize_config(v)
        else:
            normalized[k] = v
    return normalized


def get_value(obj, attr, default=None):
    """Gets a copy of the value from a nested dictionary of an object with nested attributes.

    Parameters
    ----------
    obj :
        An object or a dictionary
    attr :
        Attributes as a string seprated by dot(.)

    Returns
    -------
    Any:
        A copy of the attribute value. For dict or list, a deepcopy will be returned.

    """
    keys = attr.split(".")
    val = default
    for key in keys:
        if hasattr(obj, key):
            val = getattr(obj, key)
        elif hasattr(obj, "get"):
            val = obj.get(key, default)
        else:
            break
        obj = val
    return copy.deepcopy(val)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import logging
import time
import sys
import warnings
from abc import ABC, abstractmethod, abstractproperty
import math
import pandas as pd
import numpy as np
from scipy.stats import sem
from sklearn import set_config
from sklearn.dummy import DummyClassifier, DummyRegressor

import matplotlib.pyplot as plt

import ads
from ads.common.utils import (
    ml_task_types,
    wrap_lines,
    is_documentation_mode,
    is_notebook,
)
from ads.dataset.label_encoder import DataFrameLabelEncoder

from IPython.core.display import display, HTML

from ads.common import logger, utils


class AutoMLProvider(ABC):
    """
    Abstract Base Class defining the structure of an AutoML solution. The solution needs to
    implement train() and get_transformer_pipeline().
    """

    def __init__(self):
        self.X_train = None
        self.y_train = None
        self.X_valid = None
        self.y_valid = None
        self.client = None
        self.ml_task_type = None
        self.class_names = None
        self.transformer_pipeline = None
        self.est = None

    def setup(
        self,
        X_train,
        y_train,
        ml_task_type,
        X_valid=None,
        y_valid=None,
        class_names=None,
        client=None,
    ):
        """
        Setup arguments to the AutoML instance.

        Parameters
        ----------
        X_train : DataFrame
            Training features
        y_train : DataFrame
            Training labels
        ml_task_type : One of ml_task_type.{REGRESSION,BINARY_CLASSIFICATION,
            MULTI_CLASS_CLASSIFICATION,BINARY_TEXT_CLASSIFICATION,MULTI_CLASS_TEXT_CLASSIFICATION}
        X_valid : DataFrame
            Validation features
        y_valid : DataFrame
            Validation labels
        class_names : list
            Unique values in y_train
        client : object
            Dask client instance for distributed execution
        """
        self.X_train = X_train
        self.y_train = y_train
        self.X_valid = X_valid
        self.y_valid = y_valid
        self.ml_task_type = ml_task_type
        self.client = client
        self.class_names = class_names

    @property
    def est(self):
        """
        Returns the estimator.

        The estimator can be a standard sklearn estimator or any object that implement methods from
        (BaseEstimator, RegressorMixin) for regression or (BaseEstimator, ClassifierMixin) for classification.

        Returns
        -------
        est : An instance of estimator
        """
        return self.__est

    @est.setter
    def est(self, est):
        self.__est = est

    @abstractmethod
    def train(self, **kwargs):
        """
        Calls fit on estimator.

        This method is expected to set the 'est' property.

        Parameters
        ----------
        kwargs: dict, optional
        kwargs to decide the estimator and arguments for the fit method
        """
        pass

    @abstractmethod
    def get_transformer_pipeline(self):
        """
        Returns a list of transformers representing the transformations done on data before model prediction.

        This method is optional to implement, and is used only for visualizing transformations on data using
        ADSModel#visualize_transforms().

        Returns
        -------
        transformers_list : list of transformers implementing fit and transform
        """
        pass


class BaselineModel(object):
    """
    A BaselineModel object that supports fit/predict/predict_proba/transform
    interface. Labels (y) are encoded using DataFrameLabelEncoder.
    """

    def __init__(self, est):
        self.est = est
        self.df_label_encoder = DataFrameLabelEncoder()

    def predict(self, X):

        """
        Runs the Baselines predict function and returns the result.

        Parameters
        ----------
        X: Dataframe or list-like
          A Dataframe or list-like object holding data to be predicted on

        Returns
        -------
        List: A list of predictions performed on the input data.
        """

        X = self.transform(X)
        return self.est.predict(X)

    def predict_proba(self, X):

        """
        Runs the Baselines predict_proba function and returns the result.

        Parameters
        ----------
        X: Dataframe or list-like
          A Dataframe or list-like object holding data to be predicted on

        Returns
        -------
        List: A list of probabilities of being part of a class
        """

        X = self.transform(X)
        return self.est.predict_proba(X)

    def fit(self, X, y):

        """
        Fits the baseline estimator.

        Parameters
        ----------
        X: Dataframe or list-like
          A Dataframe or list-like object holding data to be predicted on
        Y: Dataframe, Series, or list-like
          A Dataframe, series, or list-like object holding the labels


        Returns
        -------
        estimator: The fitted estimator
        """

        self.est.fit(X, y)
        return self

    def transform(self, X):

        """
        Runs the Baselines transform function and returns the result.

        Parameters
        ---------
        X: Dataframe or list-like
          A Dataframe or list-like object holding data to be transformed

        Returns
        -------
        Dataframe or list-like: The transformed Dataframe. Currently, no transformation is performed by the default Baseline Estimator.
        """

        return X

    def __getattr__(self, item):
        return getattr(self.est, item)

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__ = state

    def __repr__(self):
        set_config()
        return str(self.est)[:-2]


class BaselineAutoMLProvider(AutoMLProvider):
    def get_transformer_pipeline(self):
        """
        Returns a list of transformers representing the transformations done on data before model prediction.

        This method is used only for visualizing transformations on data using
        ADSModel#visualize_transforms().

        Returns
        -------
        transformers_list : list of transformers implementing fit and transform
        """
        msg = "Baseline"
        return [("automl_preprocessing", AutoMLPreprocessingTransformer(msg))]

    def __init__(self, est):
        """
        Generates a baseline model using the Zero Rule algorithm by default. For a classification
        predictive modeling problem where a categorical value is predicted, the Zero
        Rule algorithm predicts the class value that has the most observations in the training dataset.

        Parameters
        ----------
        est : BaselineModel
            An estimator that supports the fit/predict/predict_proba interface.
            By default, DummyClassifier/DummyRegressor are used as estimators
        """
        super(BaselineAutoMLProvider, self).__init__()
        self.est = est

    def __repr__(self):
        set_config()
        return str(self.est)[:-2]

    def train(self, **kwargs):
        self.est = self.decide_estimator(**kwargs)
        if self.est is None:
            raise ValueError(
                "Baseline model for (%s) is not supported" % self.ml_task_type
            )
        try:
            self.est.fit(self.X_train, self.y_train)
        except Exception as e:
            warning_message = f"The baseline estimator failed to fit the data. It could not evaluate {self.est} and gave the exception {e}."
            logger.warning(warning_message)

    def decide_estimator(self, **kwargs):
        """
        Decides which type of BaselineModel to generate.

        Returns
        -------
        Modell: BaselineModel
            A baseline model generated for the particular ML task being performed
        """
        if self.est is not None:
            return self.est
        else:
            if self.ml_task_type == ml_task_types.REGRESSION:
                return BaselineModel(DummyRegressor())
            elif self.ml_task_type in [
                ml_task_types.BINARY_CLASSIFICATION,
                ml_task_types.MULTI_CLASS_CLASSIFICATION,
                ml_task_types.BINARY_TEXT_CLASSIFICATION,
                ml_task_types.MULTI_CLASS_TEXT_CLASSIFICATION,
            ]:
                return BaselineModel(DummyClassifier())


# An installation of oracle labs automl is required only for this class
class OracleAutoMLProvider(AutoMLProvider, ABC):
    def __init__(self, n_jobs=-1, loglevel=None, logger_override=None):
        """
        The Oracle AutoML Provider automatically provides a tuned ML pipeline that best models the given a training
        dataset and a prediction task at hand.

        Parameters
        ----------
        n_jobs : int
            Specifies the degree of parallelism for Oracle AutoML. -1 (default) means that AutoML will use all
            available cores.
        loglevel : int
            The verbosity of output for Oracle AutoML. Can be specified using the Python logging module
            (https://docs.python.org/3/library/logging.html#logging-levels).
        """
        try:
            self.automl = __import__("automl")
            self.cpuinfo = __import__("cpuinfo")
        except ModuleNotFoundError as e:
            utils._log_missing_module("automl", "ads[labs]")
            raise e
        super(OracleAutoMLProvider, self).__init__()
        if loglevel is None:
            loglevel = logging.DEBUG if ads.debug_mode else logging.ERROR
        else:
            loglevel = loglevel
        if logger_override:
            logr = logger_override
        else:
            logr = logging.getLogger(__name__)
        if "AMD" in self.cpuinfo.get_cpu_info().get("brand", "UNKNOWN-BRAND"):
            # Disable intra-model parallelism for LightGBM and XGBoost libraries
            # which seem to be unstable currently on AMD shapes
            self.automl.init(
                engine="local",
                engine_opts={"n_jobs": n_jobs, "model_n_jobs": 1},
                loglevel=loglevel,
            )
        else:
            self.automl.init(
                engine="local", engine_opts={"n_jobs": n_jobs}, logger=logr
            )

    def __repr__(self):
        super(OracleAutoMLProvider, self).__repr__()

    def get_transformer_pipeline(self):
        """
        Returns a list of transformers representing the transformations done on data before model prediction.

        This method is used only for visualizing transformations on data using
        ADSModel#visualize_transforms().

        Returns
        -------
        transformers_list : list of transformers implementing fit and transform
        """
        if hasattr(self.est, "text") and not self.est.text:
            msg1 = wrap_lines(
                self.est.selected_features_names_, heading="Select features:"
            )
            return [("automl_feature_selection", AutoMLFeatureSelection(msg1))]
        else:
            msg = "Apply Tfidf Vectorization\n"
            msg += "Normalize features\n"
            msg += "Label encode target"
            return [("automl_preprocessing", AutoMLPreprocessingTransformer(msg))]

    def selected_model_name(self):
        """
        Return the name of the selected model by AutoML.
        """
        return self.est.selected_model_

    def print_summary(
        self,
        max_rows=None,
        sort_column="Mean Validation Score",
        ranking_table_only=False,
    ):
        """
        Prints a summary of the Oracle AutoML Pipeline in the last train() call.

        Parameters
        ----------
        max_rows : int
            Number of trials to print. Pass in None to print all trials
        sort_column: string
            Column to sort results by. Must be one of ['Algorithm', '#Samples', '#Features', 'Mean Validation Score',
            'Hyperparameters', 'All Validation Scores', 'CPU Time']
        ranking_table_only: bool
            Table to be displayed. Pass in False to display the complete table.
            Pass in True to display the ranking table only.

        """
        if is_notebook():  # pragma: no cover
            logger.info(
                f"Training time was ({(time.time() - self.train_start_time):.2f} seconds.)"
            )

            if len(self.est.tuning_trials_) == 0 or len(self.est.train_shape_) == 0:
                logger.error(
                    "Unfortunately, there were no trials found, so we cannot visualize it."
                )
                return

            info = [
                ["Training Dataset size", self.X_train.shape],
                [
                    "Validation Dataset size",
                    self.X_valid.shape if self.X_valid is not None else None,
                ],
                ["CV", self.est.num_cv_folds_],
                ["Target variable", self.y_train.name],
                ["Optimization Metric", self.est.inferred_score_metric],
                ["Initial number of Features", self.est.train_shape_[1]],
                ["Selected number of Features", len(self.est.selected_features_names_)],
                ["Selected Features", self.est.selected_features_names_],
                ["Selected Algorithm", self.est.selected_model_],
                [
                    "End-to-end Elapsed Time (seconds)",
                    self.train_end_time - self.train_start_time,
                ],
                ["Selected Hyperparameters", self.est.selected_model_params_],
                ["Mean Validation Score", self.est.tuning_trials_[0][3]],
                ["AutoML n_jobs", self.est.n_jobs_],
                ["AutoML version", self.automl.__version__],
                ["Python version", sys.version],
            ]
            info_df = pd.DataFrame(info)

            # Remove the selected model and its params from the trials since it already shows up in the summary table
            all_trials_ = (
                self.est.model_selection_trials_
                + self.est.adaptive_sampling_trials_
                + self.est.feature_selection_trials_
                + self.est.tuning_trials_[1:]
            )
            col_names = [
                "Algorithm",
                "#Samples",
                "#Features",
                "Mean Validation Score",
                "Hyperparameters",
                "All Validation Scores",
                "CPU Time",
                "Memory Usage",
            ]
            if ranking_table_only:
                dropped_cols = [
                    "#Samples",
                    "#Features",
                    "All Validation Scores",
                    "CPU Time",
                ]
            else:
                dropped_cols = "All Validation Scores"
            summary_df = pd.DataFrame(all_trials_, columns=col_names).drop(
                dropped_cols, axis=1
            )
            sorted_summary_df = summary_df.sort_values(sort_column, ascending=False)
            # Add a rank column at the front
            sorted_summary_df.insert(
                0, "Rank based on Performance", np.arange(2, len(sorted_summary_df) + 2)
            )

            with pd.option_context(
                "display.max_colwidth",
                1000,
                "display.width",
                None,
                "display.precision",
                4,
            ):
                display(HTML(info_df.to_html(index=False, header=False)))
                if max_rows is None:
                    display(HTML(sorted_summary_df.to_html(index=False)))
                else:
                    display(
                        HTML(sorted_summary_df.to_html(index=False, max_rows=max_rows))
                    )

    def train(self, **kwargs):
        """
        Train the Oracle AutoML Pipeline. This looks at the training data, and
        identifies the best set of features, the best algorithm and the best
        set of hyperparameters for this data. A model is then generated, trained
        on this data and returned.

        Parameters
        ----------
        score_metric : str, callable
            Score function (or loss function) with signature ``score_func(y, y_pred, **kwargs)`` or string specified as
            https://scikit-learn.org/stable/modules/model_evaluation.html#common-cases-predefined-values
        random_state : int
            Random seed used by AutoML
        model_list : list of str
            Models that will be evaluated by the Pipeline. Supported models:
            - Classification: AdaBoostClassifier, DecisionTreeClassifier,
            ExtraTreesClassifier, KNeighborsClassifier,
            LGBMClassifier, LinearSVC, LogisticRegression,
            RandomForestClassifier, SVC, XGBClassifier
            - Regression: AdaBoostRegressor, DecisionTreeRegressor,
            ExtraTreesRegressor, KNeighborsRegressor,
            LGBMRegressor, LinearSVR, LinearRegression, RandomForestRegressor,
            SVR, XGBRegressor
        time_budget : float, optional
            Time budget in seconds where 0 means no time budget constraint (best effort)
        min_features : int, float, list, optional (default: 1)
            Minimum number of features to keep. Acceptable values:
            - If int, 0 < min_features <= n_features
            - If float, 0 < min_features <= 1.0
            - If list, names of features to keep, for example ['a', 'b'] means keep features 'a' and 'b'

        Returns
        -------
        self : object
        """

        """Adding this part to give the correct error for situations when dataset > 10000 rows and user tries SVC or KNN"""
        if len(self.X_train) > 10000:
            if "model_list" in kwargs:
                bad_model_list = ["SVC", "KNeighborsClassifier"]
                for model in kwargs["model_list"]:
                    for item in bad_model_list:
                        if item in model:
                            raise ValueError(
                                "SVC, KNeighborsClassifier are disabled for datasets with > 10K samples"
                            )

        self.train_start_time = time.time()
        if "time_budget" in kwargs:
            self.time_budget = kwargs.pop("time_budget")
        else:
            self.time_budget = 0  # unlimited
        self.est = self._decide_estimator(**kwargs)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.est.fit(
                self.X_train,
                self.y_train,
                X_valid=self.X_valid,
                y_valid=self.y_valid,
                time_budget=self.time_budget,
            )
        self.train_end_time = time.time()
        self.print_summary(max_rows=10)

    def print_trials(self, max_rows=None, sort_column="Mean Validation Score"):
        """
        Prints all trials executed by the Oracle AutoML Pipeline in the last train() call.

        Parameters
        ----------
        max_rows : int
            Number of trials to print. Pass in None to print all trials
        sort_column: string
            Column to sort results by. Must be one of ['Algorithm', '#Samples', '#Features', 'Mean Validation Score',
            'Hyperparameters', 'All Validation Scores', 'CPU Time']

        """
        self.est.print_trials(max_rows=max_rows, sort_column=sort_column)

    def _decide_estimator(self, **kwargs):
        """
        Decide arguments to the Oracle AutoML pipeline based on user provided
        arguments
        """
        est = None
        score_metric = None
        # Explicity define the default AutoML metrics
        if (
            self.ml_task_type == ml_task_types.BINARY_CLASSIFICATION
            or self.ml_task_type == ml_task_types.BINARY_TEXT_CLASSIFICATION
        ):
            test_model_list = ["LogisticRegression"]
        elif (
            self.ml_task_type == ml_task_types.MULTI_CLASS_CLASSIFICATION
            or self.ml_task_type == ml_task_types.MULTI_CLASS_TEXT_CLASSIFICATION
        ):
            test_model_list = ["LogisticRegression"]
        elif self.ml_task_type == ml_task_types.REGRESSION:
            test_model_list = ["LinearRegression"]
        else:
            raise ValueError("AutoML for (%s) is not supported" % self.ml_task_type)

        # Respect the user provided scoring metric if given
        if "score_metric" in kwargs:
            score_metric = kwargs.pop("score_metric")

        #
        # ***FOR TESTING PURPOSE ONLY***
        #
        # Ignore model_list for test mode
        if ads.test_mode:  # pragma: no cover
            if "model_list" in kwargs:
                _ = kwargs.pop("model_list")
            kwargs["model_list"] = test_model_list

        if (
            self.ml_task_type == ml_task_types.BINARY_CLASSIFICATION
            or self.ml_task_type == ml_task_types.MULTI_CLASS_CLASSIFICATION
        ):
            est = self.automl.Pipeline(
                task="classification", score_metric=score_metric, **kwargs
            )
        elif (
            self.ml_task_type == ml_task_types.BINARY_TEXT_CLASSIFICATION
            or self.ml_task_type == ml_task_types.MULTI_CLASS_TEXT_CLASSIFICATION
        ):
            est = self.automl.Pipeline(
                task="classification", text=True, score_metric=score_metric, **kwargs
            )
        elif self.ml_task_type == ml_task_types.REGRESSION:
            est = self.automl.Pipeline(
                task="regression", score_metric=score_metric, **kwargs
            )
        else:
            raise ValueError("AutoML for (%s) is not supported" % self.ml_task_type)
        return est

    def selected_score_label(self):
        """
        Return the name of score_metric used in train.

        """
        score_label = self.est.score_metric
        if score_label is None:
            score_label = self.est.inferred_score_metric
        return score_label

    def visualize_algorithm_selection_trials(self, ylabel=None):
        """
        Plot the scores predicted by Algorithm Selection for each algorithm. The
        horizontal line shows the average score across all algorithms. Algorithms
        below the line are colored turquoise, whereas those with a score higher
        than the mean are colored teal. The orange bar shows the algorithm with
        the highest predicted score. The error bar is +/- one standard error.

        Parameters
        ----------
        ylabel : str,
            Label for the y-axis. Defaults to the scoring metric.
        """
        if ylabel is None:
            ylabel = self.selected_score_label().capitalize()
        trials = self.est.model_selection_trials_
        if not len(trials):
            _log_visualize_no_trials("algorithm selection")
            return
        fig, ax = plt.subplots(1, figsize=(6, 3))
        colors = []
        y_error = []
        mean_scores, models, cvscores = [], [], []
        for (
            algorithm,
            samples,
            features,
            mean_score,
            hyperparameters,
            all_scores,
            runtime,
            x,
        ) in trials:
            mean_scores.append(mean_score)
            models.append(algorithm)
            cvscores.append(all_scores)
        mean_scores_ser = pd.Series(mean_scores, index=models).sort_values(
            ascending=False
        )
        scores_ser = pd.Series(cvscores, index=models)
        ax.set_title("Algorithm Selection Trials")
        ax.set_ylabel(ylabel)
        ax.set_xlabel("Algorithm")
        for f in mean_scores_ser.keys():
            se = sem(scores_ser[f], ddof=1)
            y_error.append(se)
            if f == "{}_AS".format(self.est.selected_model_):
                colors.append("orange")
            elif mean_scores_ser[f] >= mean_scores_ser.mean():
                colors.append("teal")
            else:
                colors.append("turquoise")
        mean_scores_ser.plot.bar(ax=ax, color=colors, edgecolor="black", zorder=1)
        ax.errorbar(
            x=mean_scores_ser.index.values,
            y=mean_scores_ser.values,
            yerr=y_error,
            fmt="none",
            capsize=4,
            color="black",
            zorder=0,
        )
        ax.axhline(y=mean_scores_ser.mean(), color="black", linewidth=0.5)
        ax.autoscale_view()
        plt.show()

    def visualize_adaptive_sampling_trials(self):
        """
        Visualize the trials for Adaptive Sampling.
        """
        trials = self.est.adaptive_sampling_trials_
        if len(trials) == 0:
            _log_visualize_no_trials("adaptive sampling")
            return
        fig, ax = plt.subplots(1, figsize=(6, 3))
        ax.set_title("Adaptive Sampling ({})".format(trials[0][0]))
        ax.set_xlabel("Dataset sample size")
        ax.set_ylabel(r"Predicted model score")
        scores = [
            mean_score
            for (
                algorithm,
                samples,
                features,
                mean_score,
                hyperparameters,
                all_scores,
                runtime,
                x,
            ) in trials
        ]
        n_samples = [
            samples
            for (
                algorithm,
                samples,
                features,
                mean_score,
                hyperparameters,
                all_scores,
                runtime,
                x,
            ) in trials
        ]
        y_margin = 0.10 * (max(scores) - min(scores))
        ax.grid(color="g", linestyle="-", linewidth=0.1)
        ax.set_ylim(min(scores) - y_margin, max(scores) + y_margin)
        ax.plot(n_samples, scores, "k:", marker="s", color="teal", markersize=3)
        plt.show()

    def visualize_feature_selection_trials(self, ylabel=None):
        """
        Visualize the feature selection trials taken to arrive at optimal set of
        features. The orange line shows the optimal number of features chosen
        by Feature Selection.

        Parameters
        ----------
        ylabel : str,
            Label for the y-axis. Defaults to the scoring metric.
        """
        if ylabel is None:
            ylabel = self.selected_score_label().capitalize()
        trials = self.est.feature_selection_trials_
        if len(trials) == 0:
            _log_visualize_no_trials("feature selection")
            return
        fig, ax = plt.subplots(1, figsize=(6, 3))
        ax.set_title("Feature Selection Trials")
        ax.set_xlabel("Number of Features")
        ax.set_ylabel(ylabel)
        scores = [
            mean_score
            for (
                algorithm,
                samples,
                features,
                mean_score,
                hyperparameters,
                all_scores,
                runtime,
                x,
            ) in trials
        ]
        n_features = [
            features
            for (
                algorithm,
                samples,
                features,
                mean_score,
                hyperparameters,
                all_scores,
                runtime,
                x,
            ) in trials
        ]
        y_margin = 0.10 * (max(scores) - min(scores))
        ax.grid(color="g", linestyle="-", linewidth=0.1)
        ax.set_ylim(min(scores) - y_margin, max(scores) + y_margin)
        ax.plot(n_features, scores, "k:", marker="s", color="teal", markersize=3)
        ax.axvline(
            x=len(self.est.selected_features_names_), color="orange", linewidth=2.0
        )
        plt.show()

    def visualize_tuning_trials(self, ylabel=None):
        """
        Visualize (plot) the hyperparamter tuning trials taken to arrive at the optimal
        hyper parameters. Each trial in the plot represents a particular
        hyperparamter combination.

        Parameters
        ----------
        ylabel : str,
            Label for the y-axis. Defaults to the scoring metric.
        """
        if ylabel is None:
            ylabel = self.selected_score_label().capitalize()
        # scores in trials are sorted decreasingly.
        # reversed(trails) : let the scores sort in increasing order from left to right.
        scores = [
            mean_score
            for (
                algorithm,
                samples,
                features,
                mean_score,
                hyperparameters,
                all_scores,
                runtime,
                x,
            ) in reversed(self.est.tuning_trials_)
            if mean_score and not np.isnan(mean_score)
        ]
        if not len(scores) > 1:
            raise RuntimeError("Insufficient tuning trials.")
        else:
            fig, ax = plt.subplots(1, figsize=(6, 3))
            ax.set_title("Hyperparameter Tuning Trials")
            ax.set_xlabel("Iteration $n$")
            ax.set_ylabel(ylabel)
            y_margin = 0.10 * (max(scores) - min(scores))
            ax.grid(color="g", linestyle="-", linewidth=0.1)
            ax.set_ylim(min(scores) - y_margin, max(scores) + y_margin)
            ax.plot(
                range(1, len(scores) + 1),
                scores,
                "k:",
                marker="s",
                color="teal",
                markersize=3,
            )
            plt.show()


class AutoMLPreprocessingTransformer(object):  # pragma: no cover
    def __init__(self, msg):
        self.msg = msg

    def fit(self, X):
        """
        Fits the preprocessing Transformer

        Parameters
        ----------
        X: Dataframe or list-like
          A Dataframe or list-like object holding data to be predicted on

        Returns
        -------
        Self: Estimator
            The fitted estimator
        """
        return self

    def transform(self, X):
        """
        Runs the preprocessing transform function and returns the result

        Parameters
        ---------
        X: Dataframe or list-like
          A Dataframe or list-like object holding data to be transformed

        Returns
        -------
        X: Dataframe or list-like
            The transformed Dataframe.
        """
        return X

    def _log_visualize_no_trials(target):
        logger.error(
            f"There are no trials. Therefore, the {target} cannot be visualized."
        )

    def __repr__(self):
        return self.msg


class AutoMLFeatureSelection(object):  # pragma: no cover
    def __init__(self, msg):
        self.msg = msg

    def fit(self, X):
        """
        Fits the baseline estimator

        Parameters
        ----------
        X: Dataframe or list-like
          A Dataframe or list-like object holding data to be predicted on

        Returns
        -------
        Self: Estimator
            The fitted estimator
        """
        return self

    def transform(self, X):
        """
        Runs the Baselines transform function and returns the result

        Parameters
        ---------
        X: Dataframe or list-like
          A Dataframe or list-like object holding data to be transformed

        Returns
        -------
        X: Dataframe or list-like
            The transformed Dataframe.
        """
        return X

    def __repr__(self):
        return self.msg



import warnings


def deprecated(deprecated_in, removed_in=None, details=None):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.

    Parameters
    ----------
    deprecated_in: `str`
        Version of ADS where this function deprecated.
    removed_in: `str`
        Future version where this function will be removed.
    details: `str`
        More information to be shown.
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            warnings.warn(
                f"Method {func.__name__} is "
                f"deprecated in {deprecated_in} and will be "
                f"removed in {removed_in if removed_in else 'a future release'}."
                f"{'' if not details else ' ' + details}",
                DeprecationWarning,
                stacklevel=2,
            )
            result = func(*args, **kwargs)
            return result

        return wrapper

    return decorator



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
import os
import shutil
import tempfile

import fsspec
import nbformat
from nbconvert import PythonExporter, preprocessors
from traitlets.config import Config


class Artifact:
    """Represents a OCI Data Science Job artifact.
    The Artifact class is designed to add an additional processing step on runtime/source code.
    before uploading it as data science job artifact.

    A sub-class should implement the build() method to do the additional processing.
    A sub-class is designed to be used with context manager so that the temporary files are cleaned up properly.

    For example, the NotebookArtifact implements the build() method to convert the notebook to python script.
    with NotebookArtifact(runtime) as artifact:
        # The build() method will be called when entering the context manager
        # The final artifact for the job will be stored in artifact.path
        upload_artifact(artifact.path)
        # Files are cleaned up when exit or if there is an exception.

    """

    def __init__(self, source, runtime=None) -> None:
        self.source = source
        self.path = None
        self.temp_dir = None
        self.runtime = runtime

    def __str__(self) -> str:
        if self.path:
            return self.path
        return self.source

    def __enter__(self):
        self.temp_dir = tempfile.TemporaryDirectory()
        self.build()
        return self

    def __exit__(self, *exc):
        if self.temp_dir:
            self.temp_dir.cleanup()

    def build(self):
        """Builds the runtime artifact in the temporary directory.
        Subclass should implement this method to:
        1. Process the runtime
        2. Set the self.path to the final artifact path

        Raises
        ------
        NotImplementedError
            When this method is not implemented in the subclass.
        """
        raise NotImplementedError()


class NotebookArtifact(Artifact):
    """Represents a NotebookRuntime job artifact"""

    def __init__(self, source, runtime) -> None:
        super().__init__(source, runtime=runtime)

    def build(self):
        """Prepares job artifact for notebook runtime"""
        self.path = os.path.join(
            self.temp_dir.name,
            str(os.path.basename(self.source)).replace(".ipynb", ".py"),
        )
        with fsspec.open(self.source) as f:
            nb = nbformat.reads(f.read(), nbformat.NO_CONVERT)

        exporter = PythonExporter()

        if self.runtime != None and self.runtime.exclude_tag:
            c = Config()
            c.TagRemovePreprocessor.remove_cell_tags = self.runtime.exclude_tag
            trp = preprocessors.TagRemovePreprocessor(config=c)
            trp.preprocess(nb=nb, resources=None)
        source, _ = exporter.from_notebook_node(nb)

        with open(self.path, "wb") as f:
            f.write(source.encode("utf-8"))


class ScriptArtifact(Artifact):
    """Represents a ScriptRuntime job artifact"""

    def build(self):
        """Prepares job artifact for script runtime.
        If the source is a file, it will be returned as is.
        If the source is a directory, it will be compressed as a zip file.
        """
        source = os.path.abspath(os.path.expanduser(self.source))
        path, _ = os.path.splitext(source)
        basename = os.path.basename(str(path).rstrip("/"))

        # Zip the artifact if it is a directory
        if os.path.isdir(source):
            source = str(source).rstrip("/")
            # Runtime must have entrypoint if the source is a directory
            if not self.runtime.entrypoint:
                raise ValueError(
                    "Please specify entrypoint when script source is a directory."
                )
            output = os.path.join(self.temp_dir.name, basename)
            shutil.make_archive(
                output, "zip", os.path.dirname(source), base_dir=basename
            )
            self.path = output + ".zip"
            return
        # Otherwise, use the artifact directly
        self.path = source
